# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
:term:`Email Settings <email setting>` for WuttaWeb
"""

from wuttjamaican.email import EmailSetting


class feedback(EmailSetting):  # pylint: disable=invalid-name,too-few-public-methods
    """
    Sent when user submits feedback via the web app.
    """

    default_subject = "User Feedback"

    def sample_data(self):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        user = model.User(username="barney", person=person)
        return {
            "user": user,
            "user_name": str(person),
            "user_url": "#",
            "referrer": "http://example.com/",
            "client_ip": "127.0.0.1",
            "message": "This app is cool but needs a new feature.\n\nAllow me to describe...",
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Table Views
"""

import os
import sys

from alembic import command as alembic_command
from sqlalchemy_utils import get_mapper
from mako.lookup import TemplateLookup
from webhelpers2.html import HTML

from wuttjamaican.db.conf import (
    check_alembic_current,
    make_alembic_config,
    get_alembic_scriptdir,
)

from wuttaweb.views import MasterView


class AppTableView(MasterView):  # pylint: disable=abstract-method
    """
    Master view showing all tables in the :term:`app database`.

    Default route prefix is ``app_tables``.

    Notable URLs provided by this class:

    * ``/tables/app/``
    * ``/tables/app/XXX``
    """

    # pylint: disable=duplicate-code
    model_name = "app_table"
    model_title = "App Table"
    model_key = "name"
    url_prefix = "/tables/app"
    filterable = False
    sortable = True
    sort_on_backend = False
    paginated = True
    paginate_on_backend = False
    creatable = True
    editable = False
    deletable = False
    # pylint: enable=duplicate-code

    labels = {
        "name": "Table Name",
        "module_name": "Module",
        "module_file": "File",
    }

    grid_columns = [
        "name",
        "schema",
        # "row_count",
    ]

    sort_defaults = "name"

    form_fields = [
        "name",
        "schema",
        "model_name",
        "description",
        # "row_count",
        "module_name",
        "module_file",
    ]

    has_rows = True
    rows_title = "Columns"
    rows_filterable = False
    rows_sort_defaults = "sequence"
    rows_sort_on_backend = False
    rows_paginated = True
    rows_paginate_on_backend = False
    rows_viewable = False

    row_grid_columns = [
        "sequence",
        "column_name",
        "data_type",
        "nullable",
        "description",
    ]

    def normalize_table(self, table):  # pylint: disable=missing-function-docstring
        record = {
            "name": table.name,
            "schema": table.schema or "",
            # "row_count": 42,
        }

        try:
            cls = get_mapper(table).class_
        except ValueError:
            pass
        else:
            record.update(
                {
                    "model_class": cls,
                    "model_name": cls.__name__,
                    "model_name_dotted": f"{cls.__module__}.{cls.__name__}",
                    "description": (cls.__doc__ or "").strip(),
                    "module_name": cls.__module__,
                    "module_file": sys.modules[cls.__module__].__file__,
                }
            )

        return record

    def get_grid_data(  # pylint: disable=empty-docstring
        self, columns=None, session=None
    ):
        """ """
        model = self.app.model
        data = []

        for table in model.Base.metadata.tables.values():
            data.append(self.normalize_table(table))

        return data

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # nb. show more tables by default
        g.pagesize = 50

        # schema
        g.set_searchable("schema")

        # name
        g.set_searchable("name")
        g.set_link("name")

    def get_instance(  # pylint: disable=empty-docstring,arguments-differ,unused-argument
        self, **kwargs
    ):
        """ """
        if "_cached_instance" not in self.__dict__:
            model = self.app.model

            name = self.request.matchdict["name"]
            table = model.Base.metadata.tables[name]

            # nb. sometimes need the real table reference later when
            # dealing with an instance view
            data = self.normalize_table(table)
            data["table"] = table

            self.__dict__["_cached_instance"] = data

        return self.__dict__["_cached_instance"]

    def get_instance_title(self, instance):  # pylint: disable=empty-docstring
        """ """
        return instance["name"]

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")

    def get_xref_buttons(self, obj):
        """
        By default this returns a list of buttons for each
        :class:`~wuttaweb.views.master.MasterView` subclass registered
        in the app for the current table model.  Also a button to make
        a new Master View class, if permissions allow.

        See also parent method docs,
        :meth:`~wuttaweb.views.master.MasterView.get_xref_buttons()`
        """
        table = obj
        buttons = []

        # nb. we do not omit any buttons due to lack of permission
        # here.  all buttons are shown for anyone seeing this page.
        # this is for sake of clarity so admin users are aware of what
        # is *possible* within the app etc.
        master_views = self.request.registry.settings.get("wuttaweb_master_views", {})
        model_views = master_views.get(table["model_class"], [])
        for view in model_views:
            buttons.append(
                self.make_button(
                    view.get_model_title_plural(),
                    primary=True,
                    url=self.request.route_url(view.get_route_prefix()),
                    icon_left="eye",
                )
            )

        # only add "new master view" button if user has perm
        if self.request.has_perm("master_views.create"):
            # nb. separate slightly from others
            buttons.append(HTML.tag("br"))
            buttons.append(
                self.make_button(
                    "New Master View",
                    url=self.request.route_url("master_views.create"),
                    icon_left="plus",
                )
            )

        return buttons

    def get_row_grid_data(self, obj):  # pylint: disable=empty-docstring
        """ """
        table = obj
        data = []
        for i, column in enumerate(table["table"].columns, 1):
            data.append(
                {
                    "column": column,
                    "sequence": i,
                    "column_name": column.name,
                    "data_type": str(repr(column.type)),
                    "nullable": column.nullable,
                    "description": (column.doc or "").strip(),
                }
            )
        return data

    def configure_row_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_row_grid(g)

        # nb. try not to hide any columns by default
        g.pagesize = 100

        # sequence
        g.set_label("sequence", "Seq.")

        # column_name
        g.set_searchable("column_name")

        # data_type
        g.set_searchable("data_type")

        # nullable
        g.set_renderer("nullable", "boolean")

        # description
        g.set_searchable("description")
        g.set_renderer("description", self.render_column_description)

    def render_column_description(  # pylint: disable=missing-function-docstring,unused-argument
        self, column, field, value
    ):
        if not value:
            return ""

        max_length = 100
        if len(value) <= max_length:
            return value

        return HTML.tag("span", title=value, c=f"{value[:max_length]} ...")

    def get_template_context(self, context):  # pylint: disable=empty-docstring
        """ """
        if self.creating:
            model = self.app.model

            # alembic current
            context["alembic_is_current"] = check_alembic_current(self.config)

            # existing tables
            # TODO: any reason this should check grid data instead of metadata?
            unwanted = ["transaction", "transaction_meta"]
            context["existing_tables"] = [
                {"name": table}
                for table in sorted(model.Base.metadata.tables)
                if table not in unwanted and not table.endswith("_version")
            ]

            # model dir
            context["model_dir"] = os.path.dirname(model.__file__)

            # migration branch
            script = get_alembic_scriptdir(self.config)
            branch_options = self.get_migration_branch_options(script)
            context["migration_branch_options"] = branch_options
            branch = self.config.get(
                f"{self.config.appname}.alembic.default_revise_branch"
            )
            if not branch and len(branch_options) == 1:
                branch = branch_options[0]
            context["migration_branch"] = branch

        return context

    # TODO: this is effectivey duplicated in AlembicMigrationView.get_revise_branch_options()
    def get_migration_branch_options(  # pylint: disable=missing-function-docstring
        self, script
    ):
        branches = set()
        for rev in script.get_revisions(script.get_heads()):
            branches.update(rev.branch_labels)
        return sorted(branches)

    def wizard_action(self):  # pylint: disable=too-many-return-statements
        """
        AJAX view to handle various actions for the "new table" wizard.
        """
        data = self.request.json_body
        action = data.get("action", "").strip()
        try:

            # nb. cannot use match/case statement until python 3.10, but this
            # project technically still supports python 3.8
            if action == "write_model_file":
                return self.write_model_file(data)
            if action == "check_model":
                return self.check_model(data)
            if action == "write_revision_script":
                return self.write_revision_script(data)
            if action == "migrate_db":
                return self.migrate_db(data)
            if action == "check_table":
                return self.check_table(data)
            if action == "":
                return {"error": "Must specify the action to perform."}
            return {"error": f"Unknown action requested: {action}"}

        except Exception as err:  # pylint: disable=broad-exception-caught
            return {"error": f"Unexpected error occurred: {err}"}

    def write_model_file(self, data):  # pylint: disable=missing-function-docstring
        model = self.app.model
        path = data["module_file"]

        if os.path.exists(path):
            if data["overwrite"]:
                os.remove(path)
            else:
                return {"error": "File already exists"}

        for column in data["columns"]:
            if column["data_type"]["type"] == "_fk_uuid_" and column["relationship"]:
                name = column["relationship"]

                table = model.Base.metadata.tables[column["data_type"]["reference"]]
                mapper = get_mapper(table)
                reference_model = mapper.class_.__name__

                column["relationship"] = {
                    "name": name,
                    "reference_model": reference_model,
                }

        # TODO: make templates dir configurable?
        templates = [self.app.resource_path("wuttaweb:code-templates")]
        table_templates = TemplateLookup(directories=templates)

        template = table_templates.get_template("/new-table.mako")
        content = template.render(**data)
        with open(path, "wt", encoding="utf_8") as f:
            f.write(content)

        return {}

    def check_model(self, data):  # pylint: disable=missing-function-docstring
        model = self.app.model
        model_name = data["model_name"]

        if not hasattr(model, model_name):
            return {
                "problem": "class not found in app model",
                "model": model.__name__,
            }

        return {}

    def write_revision_script(self, data):  # pylint: disable=missing-function-docstring
        alembic_config = make_alembic_config(self.config)

        script = alembic_command.revision(
            alembic_config,
            autogenerate=True,
            head=f"{data['branch']}@head",
            message=data["message"],
        )

        return {"script": script.path}

    def migrate_db(  # pylint: disable=missing-function-docstring,unused-argument
        self, data
    ):
        alembic_config = make_alembic_config(self.config)
        alembic_command.upgrade(alembic_config, "heads")
        return {}

    def check_table(self, data):  # pylint: disable=missing-function-docstring
        model = self.app.model
        name = data["name"]

        table = model.Base.metadata.tables.get(name)
        if table is None:
            return {"problem": "table does not exist in app model"}

        session = self.Session()
        count = session.query(table).count()

        route_prefix = self.get_route_prefix()
        url = self.request.route_url(f"{route_prefix}.view", name=name)
        return {"url": url, "count": count}

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._apptable_defaults(config)
        cls._defaults(config)

    # pylint: disable=duplicate-code
    @classmethod
    def _apptable_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_title_plural = cls.get_model_title_plural()
        url_prefix = cls.get_url_prefix()

        # fix permission group
        config.add_wutta_permission_group(
            permission_prefix, model_title_plural, overwrite=False
        )

        # wizard actions
        config.add_route(
            f"{route_prefix}.wizard_action",
            f"{url_prefix}/new/wizard-action",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="wizard_action",
            route_name=f"{route_prefix}.wizard_action",
            renderer="json",
            permission=f"{permission_prefix}.create",
        )

    # pylint: enable=duplicate-code


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    AppTableView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "AppTableView", base["AppTableView"]
    )
    AppTableView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Form widgets

This module defines some custom widgets for use with WuttaWeb.

However for convenience it also makes other Deform widgets available
in the namespace:

* :class:`deform:deform.widget.Widget` (base class)
* :class:`deform:deform.widget.TextInputWidget`
* :class:`deform:deform.widget.TextAreaWidget`
* :class:`deform:deform.widget.PasswordWidget`
* :class:`deform:deform.widget.CheckedPasswordWidget`
* :class:`deform:deform.widget.CheckboxWidget`
* :class:`deform:deform.widget.SelectWidget`
* :class:`deform:deform.widget.CheckboxChoiceWidget`
* :class:`deform:deform.widget.DateInputWidget`
* :class:`deform:deform.widget.DateTimeInputWidget`
* :class:`deform:deform.widget.MoneyInputWidget`
"""

import datetime
import decimal
import os

import colander
import humanize
from deform.widget import (  # pylint: disable=unused-import
    Widget,
    TextInputWidget,
    TextAreaWidget,
    PasswordWidget,
    CheckedPasswordWidget,
    CheckboxWidget,
    SelectWidget,
    CheckboxChoiceWidget,
    DateInputWidget,
    DateTimeInputWidget,
    MoneyInputWidget,
)
from webhelpers2.html import HTML, tags

from wuttjamaican.conf import parse_list


class ObjectRefWidget(SelectWidget):
    """
    Widget for use with model "object reference" fields, e.g.  foreign
    key UUID => TargetModel instance.

    While you may create instances of this widget directly, it
    normally happens automatically when schema nodes of the
    :class:`~wuttaweb.forms.schema.ObjectRef` (sub)type are part of
    the form schema; via
    :meth:`~wuttaweb.forms.schema.ObjectRef.widget_maker()`.

    In readonly mode, this renders a ``<span>`` tag around the
    :attr:`model_instance` (converted to string).

    Otherwise it renders a select (dropdown) element allowing user to
    choose from available records.

    This is a subclass of :class:`deform:deform.widget.SelectWidget`
    and uses these Deform templates:

    * ``select``
    * ``readonly/objectref``

    .. attribute:: model_instance

       Reference to the model record instance, i.e. the "far side" of
       the foreign key relationship.

       .. note::

          You do not need to provide the ``model_instance`` when
          constructing the widget.  Rather, it is set automatically
          when the :class:`~wuttaweb.forms.schema.ObjectRef` type
          instance (associated with the node) is serialized.
    """

    readonly_template = "readonly/objectref"

    def __init__(self, request, *args, **kwargs):
        url = kwargs.pop("url", None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.url = url

    def get_template_values(  # pylint: disable=empty-docstring
        self, field, cstruct, kw
    ):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        # add url, only if rendering readonly
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if (
                "url" not in values
                and self.url
                and getattr(field.schema, "model_instance", None)
            ):
                values["url"] = self.url(field.schema.model_instance)

        return values


class NotesWidget(TextAreaWidget):
    """
    Widget for use with "notes" fields.

    In readonly mode, this shows the notes with a background to make
    them stand out a bit more.

    Otherwise it effectively shows a ``<textarea>`` input element.

    This is a subclass of :class:`deform:deform.widget.TextAreaWidget`
    and uses these Deform templates:

    * ``textarea``
    * ``readonly/notes``
    """

    readonly_template = "readonly/notes"


class CopyableTextWidget(Widget):  # pylint: disable=abstract-method
    """
    A readonly text widget which adds a "copy" icon/link just after
    the text.
    """

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        if not cstruct:
            return colander.null

        return HTML.tag("wutta-copyable-text", **{"text": cstruct})

    def deserialize(self, field, pstruct):  # pylint: disable=empty-docstring
        """ """
        raise NotImplementedError


class WuttaCheckboxChoiceWidget(CheckboxChoiceWidget):
    """
    Custom widget for :class:`python:set` fields.

    This is a subclass of
    :class:`deform:deform.widget.CheckboxChoiceWidget`.

    :param request: Current :term:`request` object.

    It uses these Deform templates:

    * ``checkbox_choice``
    * ``readonly/checkbox_choice``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()


class WuttaCheckedPasswordWidget(PasswordWidget):
    """
    Custom widget for password+confirmation field.

    This widget is used only for Vue 3 + Oruga, but is *not* used for
    Vue 2 + Buefy.

    This is a subclass of :class:`deform:deform.widget.PasswordWidget`
    and uses these Deform templates:

    * ``wutta_checked_password``
    """

    template = "wutta_checked_password"


class WuttaDateWidget(DateInputWidget):
    """
    Custom widget for :class:`python:datetime.date` fields.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_date()`
    for the readonly display.

    It is automatically used for SQLAlchemy mapped classes where the
    field maps to a :class:`sqlalchemy:sqlalchemy.types.Date` column.
    For other (non-mapped) date fields, or mapped datetime fields for
    which a date widget is preferred, use
    :meth:`~wuttaweb.forms.base.Form.set_widget()`.

    This is a subclass of
    :class:`deform:deform.widget.DateInputWidget` and uses these
    Deform templates:

    * ``dateinput``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly and cstruct:
            try:
                dt = datetime.date.fromisoformat(cstruct)
            except ValueError:
                dt = datetime.datetime.fromisoformat(cstruct)
            return self.app.render_date(dt)

        return super().serialize(field, cstruct, **kw)


class WuttaDateTimeWidget(DateTimeInputWidget):
    """
    Custom widget for :class:`python:datetime.datetime` fields.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_datetime()`
    for the readonly display.

    It is automatically used for SQLAlchemy mapped classes where the
    field maps to a :class:`sqlalchemy:sqlalchemy.types.DateTime`
    column.  For other (non-mapped) datetime fields, you may have to
    use it explicitly via
    :meth:`~wuttaweb.forms.base.Form.set_widget()`.

    This is a subclass of
    :class:`deform:deform.widget.DateTimeInputWidget` and uses these
    Deform templates:

    * ``datetimeinput``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if not cstruct:
                return ""
            dt = datetime.datetime.fromisoformat(cstruct)
            return self.app.render_datetime(dt, html=True)

        return super().serialize(field, cstruct, **kw)


class WuttaMoneyInputWidget(MoneyInputWidget):
    """
    Custom widget for "money" fields.  This is used by default for
    :class:`~wuttaweb.forms.schema.WuttaMoney` type nodes.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_currency()`
    for the readonly display.

    This is a subclass of
    :class:`deform:deform.widget.MoneyInputWidget` and uses these
    Deform templates:

    * ``moneyinput``

    :param request: Current :term:`request` object.

    :param scale: If this kwarg is specified, it will be passed along
       to ``render_currency()`` call.
    """

    def __init__(self, request, *args, **kwargs):
        self.scale = kwargs.pop("scale", 2)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")
            cstruct = decimal.Decimal(cstruct)
            text = self.app.render_currency(cstruct, scale=self.scale)
            return HTML.tag("span", c=[text])

        return super().serialize(field, cstruct, **kw)


class FileDownloadWidget(Widget):  # pylint: disable=abstract-method
    """
    Widget for use with :class:`~wuttaweb.forms.schema.FileDownload`
    fields.

    This only supports readonly, and shows a hyperlink to download the
    file.  Link text is the filename plus file size.

    This is a subclass of :class:`deform:deform.widget.Widget` and
    uses these Deform templates:

    * ``readonly/filedownload``

    :param request: Current :term:`request` object.

    :param url: Optional URL for hyperlink.  If not specified, file
       name/size is shown with no hyperlink.
    """

    readonly_template = "readonly/filedownload"

    # pylint: disable=duplicate-code
    def __init__(self, request, *args, **kwargs):
        self.url = kwargs.pop("url", None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    # pylint: enable=duplicate-code

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        # nb. readonly is the only way this rolls
        kw["readonly"] = True
        template = self.readonly_template

        path = cstruct or None
        if path:
            kw.setdefault("filename", os.path.basename(path))
            kw.setdefault("filesize", self.readable_size(path))
            if self.url:
                kw.setdefault("url", self.url)

        else:
            kw.setdefault("filename", None)
            kw.setdefault("filesize", None)

        kw.setdefault("url", None)
        values = self.get_template_values(field, cstruct, kw)
        return field.renderer(template, **values)

    def readable_size(self, path):  # pylint: disable=empty-docstring
        """ """
        try:
            size = os.path.getsize(path)
        except os.error:
            size = 0
        return humanize.naturalsize(size)


class GridWidget(Widget):  # pylint: disable=abstract-method
    """
    Widget for fields whose data is represented by a :term:`grid`.

    This is a subclass of :class:`deform:deform.widget.Widget` but
    does not use any Deform templates.

    This widget only supports "readonly" mode, is not editable.  It is
    merely a convenience around the grid itself, which does the heavy
    lifting.

    Instead of creating this widget directly you probably should call
    :meth:`~wuttaweb.forms.base.Form.set_grid()` on your form.

    :param request: Current :term:`request` object.

    :param grid: :class:`~wuttaweb.grids.base.Grid` instance, used to
       display the field data.
    """

    def __init__(self, request, grid, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.grid = grid

    def serialize(self, field, cstruct, **kw):
        """
        This widget simply calls
        :meth:`~wuttaweb.grids.base.Grid.render_table_element()` on
        the ``grid`` to serialize.
        """
        readonly = kw.get("readonly", self.readonly)
        if not readonly:
            raise NotImplementedError("edit not allowed for this widget")

        return self.grid.render_table_element()


class RoleRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with User
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.User.roles` field.
    This is the default widget for the
    :class:`~wuttaweb.forms.schema.RoleRefs` type.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`.
    """

    readonly_template = "readonly/rolerefs"
    session = None

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model

        # special logic when field is editable
        readonly = kw.get("readonly", self.readonly)
        if not readonly:

            # but does not apply if current user is root
            if not self.request.is_root:
                auth = self.app.get_auth_handler()
                admin = auth.get_role_administrator(self.session)

                # prune admin role from values list; it should not be
                # one of the options since current user is not admin
                values = kw.get("values", self.values)
                values = [val for val in values if val[0] != admin.uuid]
                kw["values"] = values

        else:  # readonly

            # roles
            roles = []
            if cstruct:
                for uuid in cstruct:
                    role = self.session.get(model.Role, uuid)
                    if role:
                        roles.append(role)
            kw["roles"] = sorted(roles, key=lambda r: r.name)

            # url
            def url(role):
                return self.request.route_url("roles.view", uuid=role.uuid)

            kw["url"] = url

        # default logic from here
        return super().serialize(field, cstruct, **kw)


class PermissionsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.permissions`
    field.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`.  It uses
    these Deform templates:

    * ``permissions``
    * ``readonly/permissions``
    """

    template = "permissions"
    readonly_template = "readonly/permissions"
    permissions = None

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        kw.setdefault("permissions", self.permissions)

        if "values" not in kw:
            values = []
            for group in self.permissions.values():
                for pkey, perm in group["perms"].items():
                    values.append((pkey, perm["label"]))
            kw["values"] = values

        return super().serialize(field, cstruct, **kw)


class EmailRecipientsWidget(TextAreaWidget):
    """
    Widget for :term:`email setting` recipient fields (``To``, ``Cc``,
    ``Bcc``).

    This is a subclass of
    :class:`deform:deform.widget.TextAreaWidget`.  It uses these
    Deform templates:

    * ``textarea``
    * ``readonly/email_recips``

    See also the :class:`~wuttaweb.forms.schema.EmailRecipients`
    schema type, which uses this widget.
    """

    readonly_template = "readonly/email_recips"

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            kw["recips"] = parse_list(cstruct or "")

        return super().serialize(field, cstruct, **kw)

    def deserialize(self, field, pstruct):  # pylint: disable=empty-docstring
        """ """
        if pstruct is colander.null:
            return colander.null

        values = [value for value in parse_list(pstruct) if value]
        return ", ".join(values)


class BatchIdWidget(Widget):  # pylint: disable=abstract-method
    """
    Widget for use with the
    :attr:`~wuttjamaican:wuttjamaican.db.model.batch.BatchMixin.id`
    field of a :term:`batch` model.

    This widget is "always" read-only and renders the Batch ID as
    zero-padded 8-char string
    """

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        if cstruct is colander.null:
            return colander.null

        batch_id = int(cstruct)
        return f"{batch_id:08d}"


class AlembicRevisionWidget(Widget):  # pylint: disable=missing-class-docstring
    """
    Widget to show an Alembic revision identifier, with link to view
    the revision.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        if not cstruct:
            return colander.null

        return tags.link_to(
            cstruct, self.request.route_url("alembic.migrations.view", revision=cstruct)
        )

    def deserialize(self, field, pstruct):  # pylint: disable=empty-docstring
        """ """
        raise NotImplementedError


class AlembicRevisionsWidget(Widget):
    """
    Widget to show list of Alembic revision identifiers, with links to
    view each revision.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config

    def serialize(self, field, cstruct, **kw):  # pylint: disable=empty-docstring
        """ """
        if not cstruct:
            return colander.null

        revisions = []
        for rev in self.config.parse_list(cstruct):
            revisions.append(
                tags.link_to(
                    rev, self.request.route_url("alembic.migrations.view", revision=rev)
                )
            )

        return ", ".join(revisions)

    def deserialize(self, field, pstruct):  # pylint: disable=empty-docstring
        """ """
        raise NotImplementedError



# -*- coding: utf-8; -*-

import datetime
from collections import OrderedDict
from enum import Enum
from unittest import TestCase
from unittest.mock import patch

import sqlalchemy as sa

from wuttjamaican.db.model import Base
from wuttjamaican.util import get_timezone_by_name

from wuttaweb.grids import filters as mod
from wuttaweb.testing import WebTestCase


class TestGridFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {"name": "foo1", "value": "ONE"},
            {"name": "foo2", "value": "two"},
            {"name": "foo3", "value": "ggg"},
            {"name": "foo4", "value": "ggg"},
            {"name": "foo5", "value": "ggg"},
            {"name": "foo6", "value": "six"},
            {"name": "foo7", "value": "seven"},
            {"name": "foo8", "value": "eight"},
            {"name": "foo9", "value": "nine"},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting["name"], setting["value"])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.GridFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_constructor(self):
        model = self.app.model

        # verbs is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, "verbs"))
        filtr = self.make_filter(model.Setting.name, verbs=["foo", "bar"])
        self.assertEqual(filtr.verbs, ["foo", "bar"])

        # verb is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertIsNone(filtr.verb)
        filtr = self.make_filter(model.Setting.name, verb="foo")
        self.assertEqual(filtr.verb, "foo")

        # default verb is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, "default_verb"))
        filtr = self.make_filter(model.Setting.name, default_verb="foo")
        self.assertEqual(filtr.default_verb, "foo")

    def test_repr(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.GridFilter)
        self.assertEqual(
            repr(filtr), "GridFilter(key='name', active=False, verb=None, value=None)"
        )

    def test_get_verbs(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, "verbs"))
        self.assertEqual(filtr.default_verbs, ["equal", "not_equal"])

        # by default, returns default verbs (plus 'is_any')
        self.assertEqual(filtr.get_verbs(), ["equal", "not_equal", "is_any"])

        # default verbs can be a callable
        filtr.default_verbs = lambda: ["foo", "bar"]
        self.assertEqual(filtr.get_verbs(), ["foo", "bar", "is_any"])

        # uses filtr.verbs if set
        filtr.verbs = ["is_true", "is_false"]
        self.assertEqual(filtr.get_verbs(), ["is_true", "is_false", "is_any"])

        # may add is/null verbs
        filtr = self.make_filter(
            model.Setting.name, factory=mod.AlchemyFilter, nullable=True
        )
        self.assertEqual(
            filtr.get_verbs(),
            ["equal", "not_equal", "is_null", "is_not_null", "is_any"],
        )

        # filtr.verbs can be a callable
        filtr.nullable = False
        filtr.verbs = lambda: ["baz", "blarg"]
        self.assertEqual(filtr.get_verbs(), ["baz", "blarg", "is_any"])

    def test_get_default_verb(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, "verbs"))
        self.assertEqual(filtr.default_verbs, ["equal", "not_equal"])
        self.assertEqual(filtr.get_verbs(), ["equal", "not_equal", "is_any"])

        # returns first verb by default
        self.assertEqual(filtr.get_default_verb(), "equal")

        # returns filtr.verb if set
        filtr.verb = "foo"
        self.assertEqual(filtr.get_default_verb(), "foo")

        # returns filtr.default_verb if set
        # (nb. this overrides filtr.verb since the point of this
        # method is to return the *default* verb)
        filtr.default_verb = "bar"
        self.assertEqual(filtr.get_default_verb(), "bar")

    def test_get_verb_labels(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, "verbs"))
        self.assertEqual(filtr.get_verbs(), ["equal", "not_equal", "is_any"])

        labels = filtr.get_verb_labels()
        self.assertIsInstance(labels, dict)
        self.assertEqual(labels["equal"], "equal to")
        self.assertEqual(labels["not_equal"], "not equal to")
        self.assertEqual(labels["is_any"], "is any")

    def test_get_valueless_verbs(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, "verbs"))
        self.assertEqual(filtr.get_verbs(), ["equal", "not_equal", "is_any"])

        verbs = filtr.get_valueless_verbs()
        self.assertIsInstance(verbs, list)
        self.assertIn("is_any", verbs)

    def test_set_choices(self):
        model = self.app.model

        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertEqual(filtr.choices, {})
        self.assertEqual(filtr.data_type, "string")

        class MockEnum(Enum):
            FOO = "foo"
            BAR = "bar"

        filtr.set_choices(MockEnum)
        self.assertEqual(
            filtr.choices,
            OrderedDict(
                [
                    ("FOO", "foo"),
                    ("BAR", "bar"),
                ]
            ),
        )
        self.assertEqual(filtr.data_type, "choice")

        filtr.set_choices(None)
        self.assertEqual(filtr.choices, {})
        self.assertEqual(filtr.data_type, "string")

    def test_normalize_choices(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)

        class MockEnum(Enum):
            FOO = "foo"
            BAR = "bar"

        choices = filtr.normalize_choices(MockEnum)
        self.assertEqual(
            choices,
            OrderedDict(
                [
                    ("FOO", "foo"),
                    ("BAR", "bar"),
                ]
            ),
        )

        choices = filtr.normalize_choices(
            OrderedDict(
                [
                    ("first", "1"),
                    ("second", "2"),
                ]
            )
        )
        self.assertEqual(
            choices,
            OrderedDict(
                [
                    ("first", "1"),
                    ("second", "2"),
                ]
            ),
        )

        choices = filtr.normalize_choices(
            {
                "bbb": "b",
                "aaa": "a",
            }
        )
        self.assertEqual(
            choices,
            OrderedDict(
                [
                    ("aaa", "a"),
                    ("bbb", "b"),
                ]
            ),
        )

        choices = filtr.normalize_choices(["one", "two", "three"])
        self.assertEqual(
            choices,
            OrderedDict(
                [
                    ("one", "one"),
                    ("two", "two"),
                    ("three", "three"),
                ]
            ),
        )

    def test_apply_filter(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value, factory=mod.StringAlchemyFilter)

        # default verb used as fallback
        # self.assertEqual(filtr.default_verb, 'contains')
        filtr.default_verb = "contains"
        filtr.verb = None
        with patch.object(
            filtr, "filter_contains", side_effect=lambda q, v: q
        ) as filter_contains:
            filtered_query = filtr.apply_filter(self.sample_query, value="foo")
            filter_contains.assert_called_once_with(self.sample_query, "foo")
        self.assertIsNone(filtr.verb)

        # filter verb used as fallback
        filtr.verb = "equal"
        with patch.object(
            filtr, "filter_equal", create=True, side_effect=lambda q, v: q
        ) as filter_equal:
            filtered_query = filtr.apply_filter(self.sample_query, value="foo")
            filter_equal.assert_called_once_with(self.sample_query, "foo")

        # filter value used as fallback
        filtr.verb = "contains"
        filtr.value = "blarg"
        with patch.object(
            filtr, "filter_contains", side_effect=lambda q, v: q
        ) as filter_contains:
            filtered_query = filtr.apply_filter(self.sample_query)
            filter_contains.assert_called_once_with(self.sample_query, "blarg")

        # error if invalid verb
        self.assertRaises(
            mod.VerbNotSupported,
            filtr.apply_filter,
            self.sample_query,
            verb="doesnotexist",
        )
        filtr.verbs = ["doesnotexist"]
        self.assertRaises(
            mod.VerbNotSupported,
            filtr.apply_filter,
            self.sample_query,
            verb="doesnotexist",
        )

    def test_filter_is_any(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_any(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 9)


class TestAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {"name": "foo1", "value": "ONE"},
            {"name": "foo2", "value": "two"},
            {"name": "foo3", "value": "ggg"},
            {"name": "foo4", "value": "ggg"},
            {"name": "foo5", "value": "ggg"},
            {"name": "foo6", "value": "six"},
            {"name": "foo7", "value": "seven"},
            {"name": "foo8", "value": "eight"},
            {"name": "foo9", "value": None},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting["name"], setting["value"])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.AlchemyFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_filter_equal(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for null value
        filtered_query = filtr.filter_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)

        # nb. by default, *is filtered* by empty string
        filtered_query = filtr.filter_equal(self.sample_query, "")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 0)

        # filtered by value
        filtered_query = filtr.filter_equal(self.sample_query, "ggg")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_filter_not_equal(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_not_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)

        # nb. by default, *is filtered* by empty string
        filtered_query = filtr.filter_not_equal(self.sample_query, "")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # filtered by value
        filtered_query = filtr.filter_not_equal(self.sample_query, "ggg")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

    def test_filter_is_null(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 1)

    def test_filter_is_not_null(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_not_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 8)


class TestStringAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {"name": "foo1", "value": "ONE"},
            {"name": "foo2", "value": "two"},
            {"name": "foo3", "value": "ggg"},
            {"name": "foo4", "value": "ggg"},
            {"name": "foo5", "value": "ggg"},
            {"name": "foo6", "value": "six"},
            {"name": "foo7", "value": "seven"},
            {"name": "foo8", "value": "eight"},
            {"name": "foo9", "value": "nine"},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting["name"], setting["value"])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.StringAlchemyFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_filter_contains(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_contains(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        filtered_query = filtr.filter_contains(self.sample_query, "")
        self.assertIs(filtered_query, self.sample_query)

        # filtered by value
        filtered_query = filtr.filter_contains(self.sample_query, "ggg")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_filter_does_not_contain(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_does_not_contain(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        filtered_query = filtr.filter_does_not_contain(self.sample_query, "")
        self.assertIs(filtered_query, self.sample_query)

        # filtered by value
        filtered_query = filtr.filter_does_not_contain(self.sample_query, "ggg")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)


class TestIntegerAlchemyFilter(WebTestCase):

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.IntegerAlchemyFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_coerce_value(self):
        model = self.app.model
        filtr = self.make_filter(model.Upgrade.exit_code)

        # null
        self.assertIsNone(filtr.coerce_value(None))
        self.assertIsNone(filtr.coerce_value(""))

        # typical
        self.assertEqual(filtr.coerce_value("42"), 42)
        self.assertEqual(filtr.coerce_value("-42"), -42)

        # invalid
        self.assertIsNone(filtr.coerce_value("42.12"))
        self.assertIsNone(filtr.coerce_value("bogus"))


class TestBooleanAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {"username": "alice", "prevent_edit": False, "active": True},
            {"username": "bob", "prevent_edit": True, "active": True},
            {"username": "charlie", "active": False, "prevent_edit": None},
        ]
        for user in self.sample_data:
            user = model.User(**user)
            self.session.add(user)
        self.session.commit()
        self.sample_query = self.session.query(model.User)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.BooleanAlchemyFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_get_verbs(self):
        model = self.app.model

        # bool field, not nullable
        filtr = self.make_filter(
            model.User.active, factory=mod.BooleanAlchemyFilter, nullable=False
        )
        self.assertFalse(hasattr(filtr, "verbs"))
        self.assertEqual(filtr.default_verbs, ["is_true", "is_false"])

        # by default, returns default verbs (plus 'is_any')
        self.assertEqual(filtr.get_verbs(), ["is_true", "is_false", "is_any"])

        # default verbs can be a callable
        filtr.default_verbs = lambda: ["foo", "bar"]
        self.assertEqual(filtr.get_verbs(), ["foo", "bar", "is_any"])

        # bool field, *nullable*
        filtr = self.make_filter(
            model.User.active, factory=mod.BooleanAlchemyFilter, nullable=True
        )
        self.assertFalse(hasattr(filtr, "verbs"))
        self.assertEqual(filtr.default_verbs, ["is_true", "is_false"])

        # effective verbs also include is_false_null
        self.assertEqual(
            filtr.get_verbs(),
            [
                "is_true",
                "is_false",
                "is_false_null",
                "is_null",
                "is_not_null",
                "is_any",
            ],
        )

    def test_coerce_value(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)

        self.assertIsNone(filtr.coerce_value(None))

        self.assertTrue(filtr.coerce_value(True))
        self.assertTrue(filtr.coerce_value(1))
        self.assertTrue(filtr.coerce_value("1"))

        self.assertFalse(filtr.coerce_value(False))
        self.assertFalse(filtr.coerce_value(0))
        self.assertFalse(filtr.coerce_value(""))

    def test_filter_is_true(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_true(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

    def test_filter_is_false(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_false(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 1)

    def test_filter_is_false_null(self):
        model = self.app.model
        filtr = self.make_filter(model.User.prevent_edit)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. only one account is marked with "prevent edit"
        filtered_query = filtr.filter_is_false_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)


class TheLocalThing(Base):
    __tablename__ = "the_local_thing"
    id = sa.Column(sa.Integer(), primary_key=True, autoincrement=False)
    date = sa.Column(sa.Date(), nullable=True)
    timestamp = sa.Column(sa.DateTime(), nullable=True)


class TestDateAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        self.sample_data = [
            {"id": 1, "date": datetime.date(2024, 1, 1)},
            {"id": 2, "date": datetime.date(2024, 1, 1)},
            {"id": 3, "date": datetime.date(2024, 3, 1)},
            {"id": 4, "date": datetime.date(2024, 3, 1)},
            {"id": 5, "date": None},
            {"id": 6, "date": None},
        ]

        for thing in self.sample_data:
            thing = TheLocalThing(**thing)
            self.session.add(thing)
        self.session.commit()

        self.sample_query = self.session.query(TheLocalThing)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.DateAlchemyFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_coerce_value(self):
        filtr = self.make_filter(TheLocalThing.date)

        # null value
        self.assertIsNone(filtr.coerce_value(None))

        # value as date
        value = datetime.date(2024, 1, 1)
        result = filtr.coerce_value(value)
        self.assertIs(value, result)

        # value as string
        result = filtr.coerce_value("2024-04-01")
        self.assertIsInstance(result, datetime.date)
        self.assertEqual(result, datetime.date(2024, 4, 1))

        # invalid
        result = filtr.coerce_value("thisinputisbad")
        self.assertIsNone(result)

    def test_greater_than(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_greater_than(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date
        filtered_query = filtr.filter_greater_than(
            self.sample_query, datetime.date(2024, 2, 1)
        )
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_greater_than(self.sample_query, "2024-02-01")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

    def test_greater_equal(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_greater_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date (clear of boundary)
        filtered_query = filtr.filter_greater_equal(
            self.sample_query, datetime.date(2024, 2, 1)
        )
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as date (at boundary)
        filtered_query = filtr.filter_greater_equal(
            self.sample_query, datetime.date(2024, 3, 1)
        )
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_greater_equal(self.sample_query, "2024-01-01")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

    def test_less_than(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_less_than(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date
        filtered_query = filtr.filter_less_than(
            self.sample_query, datetime.date(2024, 2, 1)
        )
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_less_than(self.sample_query, "2024-04-01")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

    def test_less_equal(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_less_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date (clear of boundary)
        filtered_query = filtr.filter_less_equal(
            self.sample_query, datetime.date(2024, 2, 1)
        )
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as date (at boundary)
        filtered_query = filtr.filter_less_equal(
            self.sample_query, datetime.date(2024, 3, 1)
        )
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

        # value as string
        filtered_query = filtr.filter_less_equal(self.sample_query, "2024-04-01")
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)


class TestDateTimeAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()
        self.tzlocal = get_timezone_by_name("America/Los_Angeles")

        self.sample_data = [
            {
                "id": 1,
                "timestamp": self.app.make_utc(
                    datetime.datetime(2024, 1, 1, 1, 20, tzinfo=self.tzlocal)
                ),
            },
            {
                "id": 2,
                "timestamp": self.app.make_utc(
                    datetime.datetime(2024, 1, 1, 23, 40, tzinfo=self.tzlocal)
                ),
            },
            {
                "id": 3,
                "timestamp": self.app.make_utc(
                    datetime.datetime(2024, 3, 1, 22, 10, tzinfo=self.tzlocal)
                ),
            },
            {
                "id": 4,
                "timestamp": self.app.make_utc(
                    datetime.datetime(2024, 3, 1, 2, 0, tzinfo=self.tzlocal)
                ),
            },
            {"id": 5, "timestamp": None},
            {"id": 6, "timestamp": None},
        ]

        for thing in self.sample_data:
            thing = TheLocalThing(**thing)
            self.session.add(thing)
        self.session.commit()

        self.sample_query = self.session.query(TheLocalThing)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop("factory", mod.DateTimeAlchemyFilter)
        kwargs["model_property"] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_coerce_value(self):
        filtr = self.make_filter(TheLocalThing.timestamp)

        # null value
        self.assertIsNone(filtr.coerce_value(None))

        # value as date
        value = datetime.date(2024, 1, 1)
        result = filtr.coerce_value(value)
        self.assertIs(value, result)

        # value as string
        result = filtr.coerce_value("2024-04-01")
        self.assertIsInstance(result, datetime.date)
        self.assertEqual(result, datetime.date(2024, 4, 1))

        # invalid
        result = filtr.coerce_value("thisinputisbad")
        self.assertIsNone(result)

    def test_equal(self):
        model = self.app.model
        with patch.object(self.app, "get_timezone", return_value=self.tzlocal):

            filtr = self.make_filter(TheLocalThing.timestamp)
            self.assertEqual(self.sample_query.count(), 6)

            # null value ignored
            filtered_query = filtr.filter_equal(self.sample_query, None)
            self.assertIs(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 6)

            # value as date
            filtered_query = filtr.filter_equal(
                self.sample_query, datetime.date(2024, 3, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

            # value as string
            filtered_query = filtr.filter_equal(self.sample_query, "2024-02-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 0)
            filtered_query = filtr.filter_equal(self.sample_query, "2024-01-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

    def test_not_equal(self):
        model = self.app.model
        with patch.object(self.app, "get_timezone", return_value=self.tzlocal):

            filtr = self.make_filter(TheLocalThing.timestamp)
            self.assertEqual(self.sample_query.count(), 6)

            # null value ignored
            filtered_query = filtr.filter_not_equal(self.sample_query, None)
            self.assertIs(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 6)

            # value as date
            filtered_query = filtr.filter_not_equal(
                self.sample_query, datetime.date(2024, 2, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 4)

            # value as string
            filtered_query = filtr.filter_not_equal(self.sample_query, "2024-02-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 4)
            filtered_query = filtr.filter_not_equal(self.sample_query, "2024-01-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

    def test_greater_than(self):
        model = self.app.model
        with patch.object(self.app, "get_timezone", return_value=self.tzlocal):

            filtr = self.make_filter(TheLocalThing.timestamp)
            self.assertEqual(self.sample_query.count(), 6)

            # null value ignored
            filtered_query = filtr.filter_greater_than(self.sample_query, None)
            self.assertIs(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 6)

            # value as date
            filtered_query = filtr.filter_greater_than(
                self.sample_query, datetime.date(2024, 2, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

            # value as string
            filtered_query = filtr.filter_greater_than(self.sample_query, "2024-02-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

    def test_greater_equal(self):
        model = self.app.model
        with patch.object(self.app, "get_timezone", return_value=self.tzlocal):

            filtr = self.make_filter(TheLocalThing.timestamp)
            self.assertEqual(self.sample_query.count(), 6)

            # null value ignored
            filtered_query = filtr.filter_greater_equal(self.sample_query, None)
            self.assertIs(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 6)

            # value as date (clear of boundary)
            filtered_query = filtr.filter_greater_equal(
                self.sample_query, datetime.date(2024, 2, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

            # value as date (at boundary)
            filtered_query = filtr.filter_greater_equal(
                self.sample_query, datetime.date(2024, 3, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

            # value as string
            filtered_query = filtr.filter_greater_equal(self.sample_query, "2024-01-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 4)

    def test_less_than(self):
        model = self.app.model
        with patch.object(self.app, "get_timezone", return_value=self.tzlocal):

            filtr = self.make_filter(TheLocalThing.timestamp)
            self.assertEqual(self.sample_query.count(), 6)

            # null value ignored
            filtered_query = filtr.filter_less_than(self.sample_query, None)
            self.assertIs(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 6)

            # value as date
            filtered_query = filtr.filter_less_than(
                self.sample_query, datetime.date(2024, 2, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

            # value as string
            filtered_query = filtr.filter_less_than(self.sample_query, "2024-04-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 4)

    def test_less_equal(self):
        model = self.app.model
        with patch.object(self.app, "get_timezone", return_value=self.tzlocal):

            filtr = self.make_filter(TheLocalThing.timestamp)
            self.assertEqual(self.sample_query.count(), 6)

            # null value ignored
            filtered_query = filtr.filter_less_equal(self.sample_query, None)
            self.assertIs(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 6)

            # value as date (clear of boundary)
            filtered_query = filtr.filter_less_equal(
                self.sample_query, datetime.date(2024, 2, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 2)

            # value as date (at boundary)
            filtered_query = filtr.filter_less_equal(
                self.sample_query, datetime.date(2024, 3, 1)
            )
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 4)

            # value as string
            filtered_query = filtr.filter_less_equal(self.sample_query, "2024-04-01")
            self.assertIsNot(filtered_query, self.sample_query)
            self.assertEqual(filtered_query.count(), 4)


class TestVerbNotSupported(TestCase):

    def test_basic(self):
        error = mod.VerbNotSupported("equal")
        self.assertEqual(str(error), "unknown filter verb not supported: equal")



# -*- coding: utf-8; -*-

import datetime
import re
from unittest.mock import patch

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.util import get_timezone_by_name

from wuttaweb import diffs as mod
from wuttaweb.testing import WebTestCase, VersionWebTestCase


class TestWebDiff(WebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.WebDiff(self.config, *args, **kwargs)

    def test_render_html(self):
        old_data = {"foo": "bar"}
        new_data = {"foo": "baz"}
        diff = self.make_diff(old_data, new_data)
        html = diff.render_html()
        self.assertIn("<table", html)
        self.assertIn("<tr>", html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertIn(f'style="background-color: {diff.old_color}"', html)
        self.assertIn("&#39;baz&#39;", html)
        self.assertIn(f'style="background-color: {diff.new_color}"', html)
        self.assertIn("</tr>", html)
        self.assertIn("</table>", html)


class TestMergeDiff(WebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.MergeDiff(self.config, *args, **kwargs)

    def test_get_final_value_attrs(self):

        # no diff
        removing = {"foo": "bar"}
        keeping = {"foo": "bar"}
        final = {"foo": "bar"}
        is_diff = False
        diff = self.make_diff(removing, keeping, final)
        attrs = diff.get_final_value_attrs(is_diff)
        self.assertEqual(attrs, {})

        # values differ
        keeping = {"foo": "baz"}
        final = {"foo": "baz"}
        is_diff = True
        diff = self.make_diff(removing, keeping, final)
        attrs = diff.get_final_value_attrs(is_diff)
        self.assertEqual(attrs, {"class_": "has-background-warning"})

    def test_render_field_row(self):

        # no diffs
        removing = {"foo": "bar"}
        keeping = {"foo": "bar"}
        final = {"foo": "bar"}
        diff = self.make_diff(removing, keeping, final)
        html = diff.render_field_row("foo")
        self.assertTrue(html.startswith("<tr>"))
        self.assertNotIn('style="padding: 0.5rem;"', html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertNotIn(f'style="background-color: {diff.old_color}"', html)
        self.assertNotIn("&#39;baz&#39;", html)
        self.assertNotIn(f'style="background-color: {diff.new_color}"', html)
        self.assertNotIn("&#39;bar,baz&#39;", html)
        self.assertNotIn('class="has-background-warning"', html)
        self.assertTrue(html.endswith("</tr>"))

        # remove vs. keep diffs, but not keep vs. final
        removing = {"foo": "bar"}
        keeping = {"foo": "baz"}
        final = {"foo": "baz"}
        diff = self.make_diff(removing, keeping, final)
        html = diff.render_field_row("foo")
        self.assertTrue(html.startswith("<tr>"))
        self.assertNotIn('style="padding: 0.5rem;"', html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertIn(f'style="background-color: {diff.old_color}"', html)
        self.assertIn("&#39;baz&#39;", html)
        self.assertIn(f'style="background-color: {diff.new_color}"', html)
        self.assertNotIn("&#39;bar,baz&#39;", html)
        self.assertNotIn('class="has-background-warning"', html)
        self.assertTrue(html.endswith("</tr>"))

        # remove vs. keep diffs, *and* keep vs. final diffs
        removing = {"foo": "bar"}
        keeping = {"foo": "baz"}
        final = {"foo": "bar,baz"}
        diff = self.make_diff(removing, keeping, final)
        html = diff.render_field_row("foo")
        self.assertTrue(html.startswith("<tr>"))
        self.assertNotIn('style="padding: 0.5rem;"', html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertIn(f'style="background-color: {diff.old_color}"', html)
        self.assertIn("&#39;baz&#39;", html)
        self.assertIn(f'style="background-color: {diff.new_color}"', html)
        self.assertIn("&#39;bar,baz&#39;", html)
        self.assertIn('class="has-background-warning"', html)
        self.assertTrue(html.endswith("</tr>"))

        # again, with cell padding
        diff.cell_padding = "0.5rem"
        html = diff.render_field_row("foo")
        self.assertTrue(html.startswith("<tr>"))
        self.assertIn('style="padding: 0.5rem"', html)
        self.assertIn("&#39;bar&#39;", html)
        self.assertIn(f"background-color: {diff.old_color}", html)
        self.assertIn("&#39;baz&#39;", html)
        self.assertIn(f"background-color: {diff.new_color}", html)
        self.assertIn("&#39;bar,baz&#39;", html)
        self.assertIn('class="has-background-warning"', html)
        self.assertTrue(html.endswith("</tr>"))


class TestVersionDiff(VersionWebTestCase):

    def make_diff(self, *args, **kwargs):
        return mod.VersionDiff(self.config, *args, **kwargs)

    def test_constructor(self):
        import sqlalchemy_continuum as continuum

        model = self.app.model
        user = model.User(username="fred")
        self.session.add(user)
        self.session.commit()
        user.username = "freddie"
        self.session.commit()
        self.session.delete(user)
        self.session.commit()

        txncls = continuum.transaction_class(model.User)
        vercls = continuum.version_class(model.User)
        versions = self.session.query(vercls).order_by(vercls.transaction_id).all()
        self.assertEqual(len(versions), 3)

        version = versions[0]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "create")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

        version = versions[1]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "update")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

        version = versions[2]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "delete")
        self.assertEqual(
            diff.fields,
            ["active", "person_uuid", "prevent_edit", "username", "uuid"],
        )

    def test_render_version_value_with_objref(self):
        import sqlalchemy_continuum as continuum

        model = self.app.model
        person = model.Person(full_name="Fred Flintstone")
        self.session.add(person)

        # create, update, delete user
        user = model.User(username="fred", person=person)
        self.session.add(user)
        self.session.commit()
        user.username = "freddie"
        self.session.commit()
        self.session.delete(user)
        self.session.commit()

        txncls = continuum.transaction_class(model.User)
        vercls = continuum.version_class(model.User)
        versions = self.session.query(vercls).order_by(vercls.transaction_id).all()
        self.assertEqual(len(versions), 3)

        # create (1st version)
        version = versions[0]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "create")
        self.assertEqual(diff.render_old_value("username"), "")
        self.assertIn("fred", diff.render_new_value("username"))
        self.assertNotIn("freddie", diff.render_new_value("username"))
        self.assertEqual(diff.render_old_value("person_uuid"), "")
        # rendered person_uuid includes display name
        html = diff.render_new_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)

        # update (2nd version)
        version = versions[1]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "update")
        self.assertIn("fred", diff.render_old_value("username"))
        self.assertNotIn("freddie", diff.render_old_value("username"))
        self.assertIn("freddie", diff.render_new_value("username"))
        # rendered person_uuid includes display name
        html = diff.render_old_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)
        html = diff.render_new_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)

        # delete (3rd version)
        version = versions[2]
        diff = self.make_diff(version)
        self.assertEqual(diff.nature, "delete")
        self.assertIn("freddie", diff.render_old_value("username"))
        self.assertEqual(diff.render_new_value("username"), "")
        # rendered person_uuid includes display name
        html = diff.render_old_value("person_uuid")
        self.assertIn(str(person.uuid), html)
        self.assertIn("Fred Flintstone", html)
        self.assertEqual(diff.render_new_value("person_uuid"), "")

    def test_render_version_value_with_datetime(self):
        import sqlalchemy_continuum as continuum

        tzlocal = get_timezone_by_name("America/Los_Angeles")
        with patch.object(self.app, "get_timezone", return_value=tzlocal):

            # make one person
            model = self.app.model
            person = model.Person(full_name="Fred Flintstone")
            self.session.add(person)
            self.session.commit()

            # get its one version record
            txncls = continuum.transaction_class(model.Person)
            vercls = continuum.version_class(model.Person)
            version = self.session.query(vercls).order_by(vercls.transaction_id).one()

            # make a diff, but we have to mock some things up for the test
            # coverage, since we don't currently have a versioned datetime
            # field in the base model.
            diff = self.make_diff(version)

            mock_column = sa.Column("timestamp", sa.DateTime())
            mock_property = orm.ColumnProperty(column=mock_column)

            class MockMapper:
                def __init__(self, mapper):
                    self.mapper = mapper

                def __getattr__(self, name):
                    if hasattr(self.mapper, name):
                        return getattr(self.mapper, name)
                    raise AttributeError(f"attr not found: {name}")

                def has_property(self, field):
                    if field == "timestamp":
                        return True
                    return self.mapper.has_property(field)

                def get_property(self, field):
                    if field == "timestamp":
                        return mock_property
                    return self.mapper.get_property(field)

            mock_mapper = MockMapper(diff.mapper)
            with patch.object(diff, "mapper", new=mock_mapper):

                # zone-aware local time
                dt = datetime.datetime(2026, 3, 14, 14, 0, tzinfo=tzlocal)
                html = diff.render_version_value(version, "timestamp", dt)
                self.assertTrue(
                    re.search(
                        r'<span style="[^"]+"><span title="[^"]+">\d{4}-\d{2}-\d{2} \d{2}:\d{2}-\d{4}</span></span>',
                        html,
                    )
                )
                self.assertTrue(
                    re.search(
                        r'<span style="[^"]+"><span title="[^"]+">2026-03-14 14:00-0700</span></span>',
                        html,
                    )
                )

                # zone-naive (presumed UTC)
                dt = datetime.datetime(2026, 3, 14, 21, 0)
                html = diff.render_version_value(version, "timestamp", dt)
                self.assertTrue(
                    re.search(
                        r'<span style="[^"]+"><span title="[^"]+">\d{4}-\d{2}-\d{2} \d{2}:\d{2}-\d{4}</span></span>',
                        html,
                    )
                )
                self.assertTrue(
                    re.search(
                        r'<span style="[^"]+"><span title="[^"]+">2026-03-14 14:00-0700</span></span>',
                        html,
                    )
                )

                # null
                html = diff.render_version_value(version, "timestamp", None)
                self.assertFalse(
                    re.search(
                        r'<span style="[^"]+"><span title="[^"]+">\d{4}-\d{2}-\d{2} \d{2}:\d{2}-\d{4}</span></span>',
                        html,
                    )
                )



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2026 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for users
"""

import sqlalchemy as sa

from wuttjamaican.db.model import User
from wuttaweb.views import MasterView
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import PersonRef, RoleRefs


class UserView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for users.

    Default route prefix is ``users``.

    Notable URLs provided by this class:

    * ``/users/``
    * ``/users/new``
    * ``/users/XXX``
    * ``/users/XXX/edit``
    * ``/users/XXX/delete``
    """

    model_class = User

    labels = {
        "api_tokens": "API Tokens",
    }

    grid_columns = [
        "username",
        "person",
        "active",
    ]

    filter_defaults = {
        "username": {"active": True},
        "active": {"active": True, "verb": "is_true"},
    }
    sort_defaults = "username"

    form_fields = [
        "username",
        "person",
        "active",
        "prevent_edit",
        "roles",
        "api_tokens",
    ]

    mergeable = True
    merge_additive_fields = ["roles"]

    def get_query(self, session=None):  # pylint: disable=empty-docstring
        """ """
        query = super().get_query(session=session)

        # nb. always join Person
        model = self.app.model
        query = query.outerjoin(model.Person)

        return query

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)
        model = self.app.model

        # never show these
        g.remove("person_uuid", "role_refs", "password")
        g.remove_filter("password")

        # username
        g.set_link("username")

        # person
        g.set_link("person")
        g.set_sorter("person", model.Person.full_name)
        g.set_filter("person", model.Person.full_name, label="Person Full Name")

    def grid_row_class(  # pylint: disable=empty-docstring,unused-argument
        self, user, data, i
    ):
        """ """
        if not user.active:
            return "has-background-warning"
        return None

    def is_editable(self, obj):  # pylint: disable=empty-docstring
        """ """
        user = obj

        # only root can edit certain users
        if user.prevent_edit and not self.request.is_root:
            return False

        return True

    def is_deletable(self, obj):  # pylint: disable=empty-docstring
        """ """
        user = obj

        # only root can delete certain users
        if user.prevent_edit and not self.request.is_root:
            return False

        return True

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)
        user = f.model_instance

        # username
        f.set_validator("username", self.unique_username)

        # person
        if self.creating or self.editing:
            f.fields.insert_after("person", "first_name")
            f.set_required("first_name", False)
            f.fields.insert_after("first_name", "last_name")
            f.set_required("last_name", False)
            f.remove("person")
            if self.editing:
                person = user.person
                if person:
                    f.set_default("first_name", person.first_name)
                    f.set_default("last_name", person.last_name)
        else:
            f.set_node("person", PersonRef(self.request))

        # password
        # nb. we must avoid 'password' as field name since
        # ColanderAlchemy wants to handle the raw/hashed value
        f.remove("password")
        # nb. no need for password field if readonly
        if self.creating or self.editing:
            # nb. use 'set_password' as field name
            f.append("set_password")
            f.set_required("set_password", False)
            f.set_widget("set_password", widgets.CheckedPasswordWidget())

        # roles
        f.append("roles")
        f.set_node("roles", RoleRefs(self.request))
        if not self.creating:
            f.set_default("roles", [role.uuid.hex for role in user.roles])

        # api_tokens
        if self.viewing and self.has_perm("manage_api_tokens"):
            f.set_grid("api_tokens", self.make_api_tokens_grid(user))
        else:
            f.remove("api_tokens")

    def unique_username(self, node, value):  # pylint: disable=empty-docstring
        """ """
        model = self.app.model
        session = self.Session()

        query = session.query(model.User).filter(model.User.username == value)

        if self.editing:
            uuid = self.request.matchdict["uuid"]
            query = query.filter(model.User.uuid != uuid)

        if query.count():
            node.raise_invalid("Username must be unique")

    def objectify(self, form):  # pylint: disable=empty-docstring
        """ """
        auth = self.app.get_auth_handler()
        data = form.validated

        # normal logic first
        user = super().objectify(form)

        # maybe update person name
        if "first_name" in form and "last_name" in form:
            first_name = data.get("first_name")
            last_name = data.get("last_name")
            if first_name or last_name:
                user.person.full_name = self.app.make_full_name(first_name, last_name)
            else:
                user.person = None

        # maybe set user password
        if "set_password" in form and data.get("set_password"):
            auth.set_user_password(user, data["set_password"])

        # update roles for user
        # TODO
        # if self.has_perm('edit_roles'):
        self.update_roles(user, form)

        return user

    def update_roles(self, user, form):  # pylint: disable=empty-docstring
        """ """
        # TODO
        # if not self.has_perm('edit_roles'):
        #     return
        data = form.validated
        if "roles" not in data:
            return

        model = self.app.model
        session = self.Session()
        auth = self.app.get_auth_handler()

        old_roles = {role.uuid for role in user.roles}
        new_roles = data["roles"]

        admin = auth.get_role_administrator(session)
        ignored = {
            auth.get_role_authenticated(session).uuid,
            auth.get_role_anonymous(session).uuid,
        }

        # add any new roles for the user, taking care to avoid certain
        # unwanted operations for built-in roles
        for uuid in new_roles:
            if uuid in ignored:
                continue
            if uuid in old_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.append(role)

        # remove any roles which were *not* specified, taking care to
        # avoid certain unwanted operations for built-in roles
        for uuid in old_roles:
            if uuid in new_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.remove(role)

    def make_api_tokens_grid(self, user):
        """
        Make and return the grid for the API Tokens field.

        This is only shown when current user has permission to manage
        API tokens for other users.

        :rtype: :class:`~wuttaweb.grids.base.Grid`
        """
        route_prefix = self.get_route_prefix()

        grid = self.make_grid(
            key=f"{route_prefix}.view.api_tokens",
            data=[self.normalize_api_token(t) for t in user.api_tokens],
            columns=[
                "description",
                "created",
            ],
            sortable=True,
            sort_on_backend=False,
            sort_defaults=[("created", "desc")],
        )

        if self.has_perm("manage_api_tokens"):

            # create token
            button = self.make_button(
                "New",
                primary=True,
                icon_left="plus",
                **{"@click": "$emit('new-token')"},
            )
            grid.add_tool(button, key="create")

            # delete token
            grid.add_action(
                "delete",
                url="#",
                icon="trash",
                link_class="has-text-danger",
                click_handler="$emit('delete-token', props.row)",
            )

        return grid

    def normalize_api_token(self, token):  # pylint: disable=empty-docstring
        """ """
        return {
            "uuid": token.uuid.hex,
            "description": token.description,
            "created": self.app.render_datetime(token.created),
        }

    def add_api_token(self):
        """
        AJAX view for adding a new user API token.

        This calls
        :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.add_api_token()`
        for the creation logic.
        """
        session = self.Session()
        auth = self.app.get_auth_handler()
        user = self.get_instance()
        data = self.request.json_body

        token = auth.add_api_token(user, data["description"])
        session.flush()
        session.refresh(token)

        result = self.normalize_api_token(token)
        result["token_string"] = token.token_string
        result["_action_url_delete"] = "#"
        return result

    def delete_api_token(self):
        """
        AJAX view for deleting a user API token.

        This calls
        :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.delete_api_token()`
        for the deletion logic.
        """
        model = self.app.model
        session = self.Session()
        auth = self.app.get_auth_handler()
        user = self.get_instance()
        data = self.request.json_body

        token = session.get(model.UserAPIToken, data["uuid"])
        if not token:
            return {"error": "API token not found"}

        if token.user is not user:
            return {"error": "API token not found"}

        auth.delete_api_token(token)
        return {}

    def merge_get_simple_fields(self):  # pylint: disable=empty-docstring
        """ """
        fields = super().merge_get_simple_fields()

        if "password" in fields:
            fields.remove("password")

        return fields

    def merge_get_additive_fields(self):  # pylint: disable=empty-docstring
        """ """
        fields = super().merge_get_additive_fields()

        if self.app.continuum_is_enabled():
            if "transaction_count" not in fields:
                fields.append("transaction_count")

        return fields

    def merge_get_data(self, obj):  # pylint: disable=empty-docstring
        """ """
        data = super().merge_get_data(obj)
        model_class = self.get_model_class()
        session = self.Session()
        user = obj

        data["roles"] = sorted([role.name for role in user.roles])

        if self.app.continuum_is_enabled():
            import sqlalchemy_continuum as continuum  # pylint: disable=import-outside-toplevel

            txncls = continuum.transaction_class(model_class)
            data["transaction_count"] = (
                session.query(txncls).filter(txncls.user == user).count()
            )

        return data

    def merge_why_not(self, removing, keeping):
        """
        This checks to ensure the *current* user is not the same as
        the "removing" user.

        See also parent method:
        :meth:`~wuttaweb.views.master.MasterView.merge_why_not()`
        """
        if removing is self.request.user:
            return "Cannot remove user who is currently logged in!"
        return None

    def merge_execute(self, removing, keeping):
        """
        The logic to merge 2 users is extended as follows:

        The "keeping" user will be assigned to all roles to which the
        "removing" user belonged.

        Any upgrades created or executed by the "removing" user will
        be updated to reference the "keeping" user instead.

        Any versioning (SQLAlchemy-Continuum) transactions created by
        the "removing" user will be updated to reference the "keeping"
        user instead.

        See also parent method:
        :meth:`~wuttaweb.views.master.MasterView.merge_execute()`
        """
        model = self.app.model
        session = self.Session()
        model_class = self.get_model_class()

        # transfer role membership
        for role in list(removing.roles):
            if role not in keeping.roles:
                keeping.roles.append(role)

        # reassign upgrade "created by"
        stmt = (
            sa.update(model.Upgrade)
            .where(model.Upgrade.created_by_uuid == removing.uuid)
            .values(created_by_uuid=keeping.uuid)
        )
        session.execute(stmt, execution_options={"synchronize_session": False})

        # reassign upgrade "executed by"
        stmt = (
            sa.update(model.Upgrade)
            .where(model.Upgrade.executed_by_uuid == removing.uuid)
            .values(executed_by_uuid=keeping.uuid)
        )
        session.execute(stmt, execution_options={"synchronize_session": False})

        # reassign continuum transactions
        if self.app.continuum_is_enabled():
            import sqlalchemy_continuum as continuum  # pylint: disable=import-outside-toplevel

            txncls = continuum.transaction_class(model_class)
            stmt = (
                sa.update(txncls)
                .where(txncls.user_id == removing.uuid)
                .values(user_id=keeping.uuid)
            )
            session.execute(stmt, execution_options={"synchronize_session": False})

        # continue default merge
        super().merge_execute(removing, keeping)

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """

        # nb. User may come from custom model
        wutta_config = config.registry.settings["wutta_config"]
        app = wutta_config.get_app()
        cls.model_class = app.model.User

        cls._user_defaults(config)
        cls._defaults(config)

    @classmethod
    def _user_defaults(cls, config):
        """
        Provide extra default configuration for the User master view.
        """
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        instance_url_prefix = cls.get_instance_url_prefix()
        model_title = cls.get_model_title()

        # manage API tokens
        config.add_wutta_permission(
            permission_prefix,
            f"{permission_prefix}.manage_api_tokens",
            f"Manage API tokens for any {model_title}",
        )
        config.add_route(
            f"{route_prefix}.add_api_token",
            f"{instance_url_prefix}/add-api-token",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="add_api_token",
            route_name=f"{route_prefix}.add_api_token",
            permission=f"{permission_prefix}.manage_api_tokens",
            renderer="json",
        )
        config.add_route(
            f"{route_prefix}.delete_api_token",
            f"{instance_url_prefix}/delete-api-token",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="delete_api_token",
            route_name=f"{route_prefix}.delete_api_token",
            permission=f"{permission_prefix}.manage_api_tokens",
            renderer="json",
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    UserView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "UserView", base["UserView"]
    )
    UserView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
See also: :ref:`wutta-webapp`
"""

import os
import sys
from typing_extensions import Annotated

import typer
from pyramid.scripts import pserve

from wuttjamaican.cli import wutta_typer


@wutta_typer.command()
def webapp(  # pylint: disable=unused-argument
    ctx: typer.Context,
    auto_reload: Annotated[
        bool,
        typer.Option("--reload", "-r", help="Auto-reload web app when files change."),
    ] = False,
):
    """
    Run the configured web app
    """
    config = ctx.parent.wutta_config

    # we'll need config file(s) to specify for web app
    if not config.files_read:
        sys.stderr.write("no config files found!\n")
        sys.exit(1)

    runner = config.get(f"{config.appname}.web.app.runner", default="pserve")
    if runner == "pserve":

        # run pserve
        argv = ["pserve", f"file+ini:{config.files_read[0]}"]
        if ctx.params["auto_reload"]:
            argv.append("--reload")
        pserve.main(argv=argv)

    elif runner == "uvicorn":

        import uvicorn  # pylint: disable=import-error,import-outside-toplevel

        # need service details from config
        spec = config.require(f"{config.appname}.web.app.spec")
        kw = {
            "host": config.get(f"{config.appname}.web.app.host", default="127.0.0.1"),
            "port": config.get_int(f"{config.appname}.web.app.port", default=8000),
            "reload": ctx.params["auto_reload"],
            "reload_dirs": config.get_list(f"{config.appname}.web.app.reload_dirs"),
            "factory": config.get_bool(
                f"{config.appname}.web.app.factory", default=False
            ),
            "interface": config.get(
                f"{config.appname}.web.app.interface", default="auto"
            ),
            "root_path": config.get(f"{config.appname}.web.app.root_path", default=""),
        }

        # also must inject our config files to env, since there is no
        # other way to specify when running via uvicorn
        os.environ["WUTTA_CONFIG_FILES"] = os.pathsep.join(config.files_read)

        # run uvicorn
        uvicorn.run(spec, **kw)

    else:
        sys.stderr.write(f"unknown web app runner: {runner}\n")
        sys.exit(2)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2026 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tools for displaying simple data diffs
"""

import sqlalchemy as sa
from sqlalchemy import orm

from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttjamaican.diffs import Diff


class WebDiff(Diff):
    """
    Simple diff class for the web app.

    This is based on the
    :class:`~wuttjamaican:wuttjamaican.diffs.Diff` class; it just
    tweaks :meth:`render_html()` to use the web template lookup
    engine.
    """

    cell_padding = None

    def render_html(self, template="/diff.mako", **kwargs):
        """
        Render the diff as HTML table.

        :param template: Name of template to render, if you need to
           override the default.

        :param \\**kwargs: Remaining kwargs are passed as context to
           the template renderer.

        :returns: HTML literal string
        """
        context = kwargs
        context["diff"] = self
        html = render(template, context)
        return HTML.literal(html)


class MergeDiff(WebDiff):
    """
    Special diff class for use when merging 2 records.  While based on
    :class:`WebDiff`, this class uses a different signature for the
    constructor.

    It shows the "removing" record, the "keeping" record, and also the
    "final" record showing the calculated result of the merge, with
    special highlighting where values would change on the kept record.

    :param config: The app :term:`config object`.

    :param removing_data: Dict of data for the "removing" record.

    :param new_data: Dict of data for the "keeping" record.

    :param new_data: Dict of "final" data for the kept record.

    :param \\**kwargs: Remaining kwargs are passed as-is to the
       :class:`WebDiff` constructor.
    """

    def __init__(self, config, removing_data, keeping_data, final_data, **kwargs):
        super().__init__(config, removing_data, keeping_data, **kwargs)
        self.removing_data = removing_data
        self.keeping_data = keeping_data
        self.final_data = final_data
        self.columns = ["field name", "removing", "keeping", "final"]

    def render_field_row(self, field):
        keep_diff = self.keeping_data.get(field) != self.removing_data.get(field)
        final_diff = self.final_data.get(field) != self.keeping_data.get(field)

        # TODO: there is a fair bit of duplication here, compared to
        # base class.  should maybe clean that up someday..

        kw = {}
        if self.cell_padding:
            kw["style"] = f"padding: {self.cell_padding}"
        td_field = HTML.tag("td", class_="field", c=field, **kw)

        td_old_value = HTML.tag(
            "td",
            c=self.render_old_value(field),
            **self.get_old_value_attrs(keep_diff),
        )

        td_new_value = HTML.tag(
            "td",
            c=self.render_new_value(field),
            **self.get_new_value_attrs(keep_diff),
        )

        td_final_value = HTML.tag(
            "td",
            c=self.render_final_value(field),
            **self.get_final_value_attrs(final_diff),
        )

        return HTML.tag("tr", c=[td_field, td_old_value, td_new_value, td_final_value])

    def get_final_value_attrs(
        self, is_diff
    ):  # pylint: disable=missing-function-docstring
        attrs = {}
        if is_diff:
            attrs["class_"] = "has-background-warning"
        return self.get_cell_attrs(**attrs)

    def render_final_value(self, field):  # pylint: disable=missing-function-docstring
        value = repr(self.final_data.get(field))
        return self.render_cell_value(value)


class VersionDiff(WebDiff):
    """
    Special diff class for use with version history views.  While
    based on :class:`WebDiff`, this class uses a different signature
    for the constructor.

    :param config: The app :term:`config object`.

    :param version: Reference to a Continuum version record object.

    :param \\**kwargs: Remaining kwargs are passed as-is to the
       :class:`WebDiff` constructor.
    """

    def __init__(self, config, version, **kwargs):
        import sqlalchemy_continuum as continuum  # pylint: disable=import-outside-toplevel
        from wutta_continuum.util import (  # pylint: disable=import-outside-toplevel
            render_operation_type,
        )

        self.version = version
        self.model_class = continuum.parent_class(type(self.version))
        self.mapper = sa.inspect(self.model_class)
        self.version_mapper = sa.inspect(type(self.version))
        self.title = kwargs.pop("title", self.model_class.__name__)

        self.operation_title = render_operation_type(self.version.operation_type)

        if "nature" not in kwargs:
            if (
                version.previous
                and version.operation_type == continuum.Operation.DELETE
            ):
                kwargs["nature"] = "delete"
            elif version.previous:
                kwargs["nature"] = "update"
            else:
                kwargs["nature"] = "create"

        if "fields" not in kwargs:
            kwargs["fields"] = self.get_default_fields()

        old_data = {}
        new_data = {}
        for field in kwargs["fields"]:
            if version.previous:
                old_data[field] = getattr(version.previous, field)
            new_data[field] = getattr(version, field)

        super().__init__(config, old_data, new_data, **kwargs)

    def get_default_fields(self):  # pylint: disable=missing-function-docstring
        fields = sorted(self.version_mapper.columns.keys())

        unwanted = [
            "transaction_id",
            "end_transaction_id",
            "operation_type",
        ]

        return [field for field in fields if field not in unwanted]

    def render_version_value(self, version, field, value):
        """
        Render the cell value HTML for a given version + field.

        This method is used to render both sides of the diff (old +
        new values).  It will just render the field value using a
        monospace font by default.  However:

        If the field is involved in a mapper relationship (i.e. it is
        the "foreign key" to a related table), the logic here will
        also (try to) traverse that show display text for the related
        object (if found).

        :param version: Reference to the Continuum version object.

        :param field: Name of the field, as string.

        :param value: Raw value for the field, as obtained from the
           version object.

        :returns: Rendered cell value as HTML literal
        """
        # first render normal span; this is our fallback but also may
        # be embedded within a more complex result.
        text = HTML.tag("span", c=[repr(value)], style="font-family: monospace;")

        # style to apply for bold human-friendly text (if applicable)
        bold = "margin-left: 2rem; font-style: italic; font-weight: bold;"

        # check for standard datetime field
        if self.mapper.has_property(field):
            prop = self.mapper.get_property(field)
            if isinstance(prop, orm.ColumnProperty):
                if len(prop.columns) == 1:
                    col = prop.columns[0]
                    if isinstance(col.type, sa.DateTime):
                        if value:
                            # render as local datetime w/ "time since" tooltip
                            display = HTML.tag(
                                "span",
                                c=self.app.render_datetime(value, html=True),
                                style=bold,
                            )
                            return HTML.tag("span", c=[text, display])

        # loop thru all mapped relationship props
        for prop in self.mapper.relationships:

            # we only want singletons
            if prop.uselist:
                continue

            # loop thru columns for prop
            # nb. there should always be just one colum for a
            # singleton prop, but technically a list is used, so no
            # harm in looping i assume..
            for col in prop.local_columns:

                # we only want the matching column
                if col.name != field:
                    continue

                # grab "related version" reference via prop key.  this
                # would be like a UserVersion for instance.
                if ref := getattr(version, prop.key):

                    # grab "related object" reference.  this would be
                    # like a User for instance.
                    if ref := getattr(ref, "version_parent", None):

                        # render text w/ related object as bold string
                        return HTML.tag(
                            "span",
                            c=[text, HTML.tag("span", c=[str(ref)], style=bold)],
                        )

        return text

    def render_old_value(self, field):
        if self.nature == "create":
            return ""
        value = self.old_value(field)
        return self.render_version_value(self.version.previous, field, value)

    def render_new_value(self, field):
        if self.nature == "delete":
            return ""
        value = self.new_value(field)
        return self.render_version_value(self.version, field, value)



# -*- coding: utf-8; -*-

from wuttaweb.menus import MenuHandler


class NullMenuHandler(MenuHandler):
    """
    Dummy :term:`menu handler` for testing.
    """

    def make_menus(self, request, **kwargs):
        """
        This always returns an empty menu set.
        """
        return []



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2026 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Grid Filters
"""

import datetime
import logging
from collections import OrderedDict

try:
    from enum import EnumType
except ImportError:  # pragma: no cover
    # nb. python <= 3.10
    from enum import EnumMeta as EnumType

import sqlalchemy as sa

from wuttjamaican.util import UNSPECIFIED


log = logging.getLogger(__name__)


class VerbNotSupported(Exception):  # pylint: disable=empty-docstring
    """ """

    def __init__(self, verb):
        self.verb = verb

    def __str__(self):
        return f"unknown filter verb not supported: {self.verb}"


class GridFilter:  # pylint: disable=too-many-instance-attributes
    """
    Filter option for a grid.  Represents both the "features" as well
    as "state" for the filter.

    :param request: Current :term:`request` object.

    :param nullable: Boolean indicating whether the filter should
       include ``is_null`` and ``is_not_null`` verbs.  If not
       specified, the column will be inspected (if possible) and use
       its nullable flag.

    :param \\**kwargs: Any additional kwargs will be set as attributes
       on the filter instance.

    Filter instances have the following attributes:

    .. attribute:: key

       Unique key for the filter.  This often corresponds to a "column
       name" for the grid, but not always.

    .. attribute:: label

       Display label for the filter field.

    .. attribute:: data_type

       Simplistic "data type" which the filter supports.  So far this
       will be one of:

       * ``'string'``
       * ``'date'``
       * ``'choice'``

       Note that this mainly applies to the "value input" used by the
       filter.  There is no data type for boolean since it does not
       need a value input; the verb is enough.

    .. attribute:: active

       Boolean indicating whether the filter is currently active.

       See also :attr:`verb` and :attr:`value`.

    .. attribute:: verb

       Verb for current filter, if :attr:`active` is true.

       See also :attr:`value`.

    .. attribute:: choices

       OrderedDict of possible values for the filter.

       This is safe to read from, but use :meth:`set_choices()` to
       update it.

    .. attribute:: value

       Value for current filter, if :attr:`active` is true.

       See also :attr:`verb`.

    .. attribute:: default_active

       Boolean indicating whether the filter should be active by
       default, i.e. when first displaying the grid.

       See also :attr:`default_verb` and :attr:`default_value`.

    .. attribute:: default_verb

       Filter verb to use by default.  This will be auto-selected when
       the filter is first activated, or when first displaying the
       grid if :attr:`default_active` is true.

       See also :attr:`default_value`.

    .. attribute:: default_value

       Filter value to use by default.  This will be auto-populated
       when the filter is first activated, or when first displaying
       the grid if :attr:`default_active` is true.

       See also :attr:`default_verb`.
    """

    data_type = "string"
    default_verbs = ["equal", "not_equal"]

    default_verb_labels = {
        "is_any": "is any",
        "equal": "equal to",
        "not_equal": "not equal to",
        "greater_than": "greater than",
        "greater_equal": "greater than or equal to",
        "less_than": "less than",
        "less_equal": "less than or equal to",
        # 'between':              "between",
        "is_true": "is true",
        "is_false": "is false",
        "is_false_null": "is false or null",
        "is_null": "is null",
        "is_not_null": "is not null",
        "contains": "contains",
        "does_not_contain": "does not contain",
    }

    valueless_verbs = [
        "is_any",
        "is_true",
        "is_false",
        "is_false_null",
        "is_null",
        "is_not_null",
    ]

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        request,
        key,
        label=None,
        verbs=None,
        choices=None,
        nullable=None,
        default_active=False,
        default_verb=None,
        default_value=None,
        **kwargs,
    ):
        self.request = request
        self.key = key
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.label = label or self.app.make_title(self.key)

        # remember original data type in case we need to revert,
        # e.g. after changing it to 'choices' and back again
        self.original_data_type = self.data_type

        # active
        self.default_active = default_active
        self.active = self.default_active

        # verb
        if verbs is not None:
            self.verbs = verbs
        if default_verb:
            self.default_verb = default_verb
        self.verb = None  # active verb is set later

        # choices
        self.set_choices(choices)

        # nullable
        self.nullable = nullable

        # value
        self.default_value = default_value
        self.value = self.default_value

        self.__dict__.update(kwargs)

    def __repr__(self):
        verb = getattr(self, "verb", None)
        return (
            f"{self.__class__.__name__}("
            f"key='{self.key}', "
            f"active={self.active}, "
            f"verb={repr(verb)}, "
            f"value={repr(self.value)})"
        )

    def get_verbs(self):
        """
        Returns the list of verbs supported by the filter.
        """
        verbs = None

        if hasattr(self, "verbs"):
            verbs = self.verbs

        else:
            verbs = self.default_verbs

        if callable(verbs):
            verbs = verbs()
        verbs = list(verbs)

        if self.nullable:
            if "is_null" not in verbs:
                verbs.append("is_null")
            if "is_not_null" not in verbs:
                verbs.append("is_not_null")

        if "is_any" not in verbs:
            verbs.append("is_any")

        return verbs

    def get_verb_labels(self):
        """
        Returns a dict of all defined verb labels.
        """
        # TODO: should traverse hierarchy
        labels = {verb: verb for verb in self.get_verbs()}
        labels.update(self.default_verb_labels)
        return labels

    def get_valueless_verbs(self):
        """
        Returns a list of verb names which do not need a value.
        """
        return self.valueless_verbs

    def get_default_verb(self):
        """
        Returns the default verb for the filter.
        """
        verb = None

        if hasattr(self, "default_verb"):
            verb = self.default_verb

        elif hasattr(self, "verb"):
            verb = self.verb

        if not verb:
            verbs = self.get_verbs()
            if verbs:
                verb = verbs[0]

        return verb

    def set_choices(self, choices):
        """
        Set the value choices for the filter.

        If ``choices`` is non-empty, it is passed to
        :meth:`normalize_choices()` and the result is assigned to
        :attr:`choices`.  Also, the :attr:`data_type` is set to
        ``'choice'`` so the UI will present the value input as a
        dropdown.

        But if ``choices`` is empty, :attr:`choices` is set to an
        empty dict, and :attr:`data_type` is set (back) to
        ``'string'``.

        :param choices: Collection of "choices" or ``None``.
        """
        if choices:
            self.choices = self.normalize_choices(choices)
            self.data_type = "choice"
        else:
            self.choices = {}
            if self.data_type == "choice":
                self.data_type = self.original_data_type

    def normalize_choices(self, choices):
        """
        Normalize a collection of "choices" to standard ``OrderedDict``.

        This is called automatically by :meth:`set_choices()`.

        :param choices: A collection of "choices" in one of the following
           formats:

           * :class:`python:enum.Enum` class
           * simple list, each value of which should be a string,
             which is assumed to be able to serve as both key and
             value (ordering of choices will be preserved)
           * simple dict, keys and values of which will define the
             choices (note that the final choices will be sorted by
             key!)
           * OrderedDict, keys and values of which will define the
             choices (ordering of choices will be preserved)

        :rtype: :class:`python:collections.OrderedDict`
        """
        normalized = choices

        if isinstance(choices, EnumType):
            normalized = OrderedDict(
                [(member.name, member.value) for member in choices]
            )

        elif isinstance(choices, OrderedDict):
            normalized = choices

        elif isinstance(choices, dict):
            normalized = OrderedDict([(key, choices[key]) for key in sorted(choices)])

        elif isinstance(choices, list):
            normalized = OrderedDict([(key, key) for key in choices])

        return normalized

    def coerce_value(self, value):
        """
        Coerce the given value to the correct type/format for use with
        the filter.  This is where e.g. a boolean or date filter
        should convert input string to ``bool`` or ``date`` value.

        This is (usually) called from a filter method, when applying
        the filter.  See also :meth:`apply_filter()`.

        Default logic on the base class returns value as-is; subclass
        may override as needed.

        :param value: Input string provided by the user via the filter
           form submission.

        :returns: Value of the appropriate type, depending on the
           filter subclass.
        """
        return value

    def apply_filter(self, data, verb=None, value=UNSPECIFIED):
        """
        Filter the given data set according to a verb/value pair.

        If verb and/or value are not specified, will use :attr:`verb`
        and/or :attr:`value` instead.

        This method does not directly filter the data; rather it
        delegates (based on ``verb``) to some other method.  The
        latter may choose *not* to filter the data, e.g. if ``value``
        is empty, in which case this may return the original data set
        unchanged.

        :returns: The (possibly) filtered data set.
        """
        if verb is None:
            verb = self.verb
        if not verb:
            verb = self.get_default_verb()
            log.warning(
                "missing verb for '%s' filter, will use default verb: %s",
                self.key,
                verb,
            )

        # only attempt for known verbs
        if verb not in self.get_verbs():
            raise VerbNotSupported(verb)

        # fallback value
        if value is UNSPECIFIED:
            value = self.value

        # locate filter method
        func = getattr(self, f"filter_{verb}", None)
        if not func:
            raise VerbNotSupported(verb)

        # invoke filter method
        return func(data, value)  # pylint: disable=not-callable

    def filter_is_any(self, data, value):  # pylint: disable=unused-argument
        """
        This is a no-op which always ignores the value and returns the
        data as-is.
        """
        return data


class AlchemyFilter(GridFilter):
    """
    Filter option for a grid with SQLAlchemy query data.

    This is a subclass of :class:`GridFilter`.  It requires a
    ``model_property`` to know how to filter the query.

    :param model_property: Property of a model class, representing the
       column by which to filter.  For instance,
       ``model.Person.full_name``.
    """

    def __init__(self, *args, **kwargs):
        self.model_property = kwargs.pop("model_property")
        super().__init__(*args, **kwargs)

        if self.nullable is None:
            columns = self.model_property.prop.columns
            if len(columns) == 1:
                self.nullable = columns[0].nullable

    def filter_equal(self, query, value):
        """
        Filter data with an equal (``=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        return query.filter(self.model_property == value)

    def filter_not_equal(self, query, value):
        """
        Filter data with a not equal (``!=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        # sql probably excludes null values from results, but user
        # probably does not expect that, so explicitly include them.
        return query.filter(
            sa.or_(
                self.model_property == None,  # pylint: disable=singleton-comparison
                self.model_property != value,
            )
        )

    def filter_greater_than(self, query, value):
        """
        Filter data with a greater than (``>``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property > value)

    def filter_greater_equal(self, query, value):
        """
        Filter data with a greater than or equal (``>=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property >= value)

    def filter_less_than(self, query, value):
        """
        Filter data with a less than (``<``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property < value)

    def filter_less_equal(self, query, value):
        """
        Filter data with a less than or equal (``<=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property <= value)

    def filter_is_null(self, query, value):  # pylint: disable=unused-argument
        """
        Filter data with an ``IS NULL`` query.  The value is ignored.
        """
        return query.filter(
            self.model_property == None  # pylint: disable=singleton-comparison
        )

    def filter_is_not_null(self, query, value):  # pylint: disable=unused-argument
        """
        Filter data with an ``IS NOT NULL`` query.  The value is
        ignored.
        """
        return query.filter(
            self.model_property != None  # pylint: disable=singleton-comparison
        )


class StringAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a text data column.

    Subclass of :class:`AlchemyFilter`.
    """

    default_verbs = ["contains", "does_not_contain", "equal", "not_equal"]

    def coerce_value(self, value):  # pylint: disable=empty-docstring
        """ """
        if value is not None:
            value = str(value)
            if value:
                return value
        return None

    def filter_contains(self, query, value):
        """
        Filter data with an ``ILIKE`` condition.
        """
        value = self.coerce_value(value)
        if not value:
            return query

        criteria = []
        for val in value.split():
            val = val.replace("_", r"\_")
            val = f"%{val}%"
            criteria.append(self.model_property.ilike(val))

        return query.filter(sa.and_(*criteria))

    def filter_does_not_contain(self, query, value):
        """
        Filter data with a ``NOT ILIKE`` condition.
        """
        value = self.coerce_value(value)
        if not value:
            return query

        criteria = []
        for val in value.split():
            val = val.replace("_", r"\_")
            val = f"%{val}%"
            criteria.append(~self.model_property.ilike(val))

        # sql probably excludes null values from results, but user
        # probably does not expect that, so explicitly include them.
        return query.filter(
            sa.or_(
                self.model_property == None,  # pylint: disable=singleton-comparison
                sa.and_(*criteria),
            )
        )


class NumericAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a numeric data column.

    Subclass of :class:`AlchemyFilter`.
    """

    default_verbs = [
        "equal",
        "not_equal",
        "greater_than",
        "greater_equal",
        "less_than",
        "less_equal",
    ]


class IntegerAlchemyFilter(NumericAlchemyFilter):
    """
    SQLAlchemy filter option for an integer data column.

    Subclass of :class:`NumericAlchemyFilter`.
    """

    def coerce_value(self, value):  # pylint: disable=empty-docstring
        """ """
        if value:
            try:
                return int(value)
            except Exception:  # pylint: disable=broad-exception-caught
                pass
        return None


class BooleanAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a boolean data column.

    Subclass of :class:`AlchemyFilter`.
    """

    default_verbs = ["is_true", "is_false"]

    def get_verbs(self):  # pylint: disable=empty-docstring
        """ """

        # get basic verbs from caller, or default list
        verbs = getattr(self, "verbs", self.default_verbs)
        if callable(verbs):
            verbs = verbs()
        verbs = list(verbs)

        # add some more if column is nullable
        if self.nullable:
            for verb in ("is_false_null", "is_null", "is_not_null"):
                if verb not in verbs:
                    verbs.append(verb)

        # add wildcard
        if "is_any" not in verbs:
            verbs.append("is_any")

        return verbs

    def coerce_value(self, value):  # pylint: disable=empty-docstring
        """ """
        if value is not None:
            return bool(value)
        return None

    def filter_is_true(self, query, value):  # pylint: disable=unused-argument
        """
        Filter data with an "is true" condition.  The value is
        ignored.
        """
        return query.filter(
            self.model_property == True  # pylint: disable=singleton-comparison
        )

    def filter_is_false(self, query, value):  # pylint: disable=unused-argument
        """
        Filter data with an "is false" condition.  The value is
        ignored.
        """
        return query.filter(
            self.model_property == False  # pylint: disable=singleton-comparison
        )

    def filter_is_false_null(self, query, value):  # pylint: disable=unused-argument
        """
        Filter data with "is false or null" condition.  The value is
        ignored.
        """
        return query.filter(
            sa.or_(
                self.model_property == False,  # pylint: disable=singleton-comparison
                self.model_property == None,  # pylint: disable=singleton-comparison
            )
        )


class DateAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a
    :class:`~sqlalchemy:sqlalchemy.types.Date` column.

    Subclass of :class:`AlchemyFilter`.

    This filter class has custom logic to coerce the input value, but
    does not have custom filter logic beyond that.
    """

    data_type = "date"
    default_verbs = [
        "equal",
        "not_equal",
        "greater_than",
        "greater_equal",
        "less_than",
        "less_equal",
        # 'between',
    ]

    default_verb_labels = {
        "equal": "on",
        "not_equal": "not on",
        "greater_than": "after",
        "greater_equal": "on or after",
        "less_than": "before",
        "less_equal": "on or before",
        # 'between':              "between",
        "is_any": "is any",
    }

    def coerce_value(self, value):
        """
        Convert the given value to a proper
        :class:`python:datetime.date` object, if applicable.

        If the input value is already a date object, it is returned
        as-is.

        Otherwise it is assumed to be a string in ``%Y-%m-%d`` format,
        and will be converted to a date object.

        If the conversion fails, or no value is provided, ``None`` is
        returned.
        """
        if value:
            if isinstance(value, datetime.date):
                return value

            try:
                dt = datetime.datetime.strptime(value, "%Y-%m-%d")
            except ValueError:
                log.warning("invalid date value: %s", value)
            else:
                return dt.date()

        return None


class DateTimeAlchemyFilter(DateAlchemyFilter):
    """
    SQLAlchemy filter option for a
    :class:`~sqlalchemy:sqlalchemy.types.DateTime` column.

    Subclass of :class:`DateAlchemyFilter`.

    This filter class has custom logic to coerce the input value,
    inherited from parent class.  It also has custom filter logic
    for most verbs (not/equal, greater/less than etc.).

    Please note that this class assumes the underlying data uses
    "naive UTC" values.  It therefore will convert to/from local time
    zone accordingly, to ensure user gets the data they expect.  For
    more info see :doc:`wuttjamaican:narr/datetime`.
    """

    def get_start_datetime(self, value, as_utc=True):
        """
        Calculate the "start" timestamp for the given date value.

        The return value will be the "first possible moment" of the
        given date.

        :param value: :class:`python:datetime.date` instance

        :param as_utc: Indicates the return value should be naive/UTC;
           set this to ``False`` to get the aware/local value.

        :returns: :class:`python:datetime.datetime` instance
        """
        start = datetime.datetime.combine(value, datetime.time(0))
        start = self.app.localtime(start, from_utc=False)
        if as_utc:
            start = self.app.make_utc(start)
        return start

    def get_end_datetime(self, value, as_utc=True):
        """
        Calculate the "end" timestamp for the given date value.

        Due to the nature of queries involving this "end" boundary,
        the return value will be the "first possible moment" of the
        day *after* the given date.

        :param value: :class:`python:datetime.date` instance

        :param as_utc: Indicates the return value should be naive/UTC;
           set this to ``False`` to get the aware/local value.

        :returns: :class:`python:datetime.datetime` instance
        """
        end = datetime.datetime.combine(
            value + datetime.timedelta(days=1), datetime.time(0)
        )
        end = self.app.localtime(end, from_utc=False)
        if as_utc:
            end = self.app.make_utc(end)
        return end

    def filter_equal(self, query, value):
        """
        Find all records with datetime values which fall on the given
        date.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        start = self.get_start_datetime(value)
        end = self.get_end_datetime(value)
        return query.filter(self.model_property >= start).filter(
            self.model_property < end
        )

    def filter_not_equal(self, query, value):
        """
        Find all records with datetime values which fall outside the
        given date.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        start = self.get_start_datetime(value)
        end = self.get_end_datetime(value)
        return query.filter(
            sa.or_(self.model_property < start, self.model_property >= end)
        )

    def filter_greater_than(self, query, value):
        """
        Find all records with datetime values which fall after the
        given date.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        end = self.get_end_datetime(value)
        return query.filter(self.model_property >= end)

    def filter_greater_equal(self, query, value):
        """
        Find all records with datetime values which fall on or after
        the given date.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        start = self.get_start_datetime(value)
        return query.filter(self.model_property >= start)

    def filter_less_than(self, query, value):
        """
        Find all records with datetime values which fall before the
        given date.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        start = self.get_start_datetime(value)
        return query.filter(self.model_property < start)

    def filter_less_equal(self, query, value):
        """
        Find all records with datetime values which fall on or before
        the given date.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        end = self.get_end_datetime(value)
        return query.filter(self.model_property < end)


default_sqlalchemy_filters = {
    None: AlchemyFilter,
    sa.String: StringAlchemyFilter,
    sa.Text: StringAlchemyFilter,
    sa.Numeric: NumericAlchemyFilter,
    sa.Integer: IntegerAlchemyFilter,
    sa.Boolean: BooleanAlchemyFilter,
    sa.Date: DateAlchemyFilter,
    sa.DateTime: DateTimeAlchemyFilter,
}



# -*- coding: utf-8; -*-

from wuttjamaican.db.model import User
from wuttjamaican.testing import ConfigTestCase

from wuttaweb import conf as mod
from wuttaweb.testing import WebTestCase
from wuttaweb.views import MasterView


class TestWuttaWebConfigExtension(ConfigTestCase):

    def test_basic(self):

        # continuum plugin not set yet (b/c config was not extended)
        self.assertIsNone(self.config.get("wutta_continuum.wutta_plugin_spec"))

        # so let's extend it
        extension = mod.WuttaWebConfigExtension()
        extension.configure(self.config)
        self.assertEqual(
            self.config.get("wutta_continuum.wutta_plugin_spec"),
            "wuttaweb.db.continuum:WuttaWebContinuumPlugin",
        )


class MasterWithClass(MasterView):
    model_class = User


class MasterWithName(MasterView):
    model_class = "Widget"


class TestAddMasterView(WebTestCase):

    def test_master_with_class(self):
        model = self.app.model

        # nb. due to minimal test bootstrapping, no master views are
        # registered by default at this point
        self.assertNotIn("wuttaweb_master_views", self.request.registry.settings)

        self.pyramid_config.add_wutta_master_view(MasterWithClass)
        self.assertIn("wuttaweb_master_views", self.request.registry.settings)
        master_views = self.request.registry.settings["wuttaweb_master_views"]
        self.assertIn(model.User, master_views)
        self.assertEqual(master_views[model.User], [MasterWithClass])

    def test_master_with_name(self):
        model = self.app.model

        # nb. due to minimal test bootstrapping, no master views are
        # registered by default at this point
        self.assertNotIn("wuttaweb_master_views", self.request.registry.settings)

        self.pyramid_config.add_wutta_master_view(MasterWithName)
        self.assertIn("wuttaweb_master_views", self.request.registry.settings)
        master_views = self.request.registry.settings["wuttaweb_master_views"]
        self.assertIn("Widget", master_views)
        self.assertEqual(master_views["Widget"], [MasterWithName])



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Config Extension
"""

from wuttjamaican.conf import WuttaConfigExtension


class WuttaWebConfigExtension(WuttaConfigExtension):
    """
    Config extension for WuttaWeb.

    This sets the default plugin used for SQLAlchemy-Continuum, to
    :class:`~wuttaweb.db.continuum.WuttaWebContinuumPlugin`.  Which is
    only relevant if Wutta-Continuum is installed and enabled.  For
    more info see :doc:`wutta-continuum:index`.
    """

    key = "wuttaweb"

    def configure(self, config):  # pylint: disable=empty-docstring
        """ """
        config.setdefault(
            "wutta_continuum.wutta_plugin_spec",
            "wuttaweb.db.continuum:WuttaWebContinuumPlugin",
        )


def add_master_view(config, master):
    """
    Pyramid directive to add the given ``MasterView`` subclass to the
    app's registry.

    This allows the app to dynamically present certain options for
    admin features etc.

    This is normally called automatically for all master views, within
    the :meth:`~wuttaweb.views.master.MasterView.defaults()` method.

    Should you need to call this yourself, do not call it directly but
    instead make a similar call via the Pyramid config object::

       pyramid_config.add_wutta_master_view(PoserWidgetView)

    :param config: Reference to the Pyramid config object.

    :param master: Reference to a
       :class:`~wuttaweb.views.master.MasterView` subclass.

    This function is involved in app startup; once that phase is
    complete you can inspect the master views like so::

       master_views = request.registry.settings["wuttaweb_master_views"]

       # find master views for given model class
       user_views = master_views.get(model.User, [])

       # some master views are registered by model name instead (if no class)
       email_views = master_views.get("email_setting", [])
    """
    key = master.get_model_class() or master.get_model_name()

    def action():
        master_views = config.get_settings().get("wuttaweb_master_views", {})
        master_views.setdefault(key, []).append(master)
        config.add_settings({"wuttaweb_master_views": master_views})

    config.action(None, action)



# -*- coding: utf-8; -*-

from unittest.mock import patch

from wuttjamaican.email import EmailSetting, EmailHandler

import colander
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response

from wuttaweb import emails as wuttaweb_emails
from wuttaweb.views import email as mod
from wuttaweb.testing import WebTestCase


class TestEmailSettingViews(WebTestCase):

    def make_view(self):
        return mod.EmailSettingView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.email")

    def test_get_grid_data(self):
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        view = self.make_view()

        # nb. avoid issues with arbitrary installed packages
        email_modules = [wuttaweb_emails]
        with patch.object(
            EmailHandler, "get_email_modules", return_value=email_modules
        ):
            data = view.get_grid_data()
            self.assertIsInstance(data, list)
            self.assertTrue(data)  # 1+ items
            setting = data[0]
            self.assertIn("key", setting)
            self.assertIn("subject", setting)
            self.assertIn("sender", setting)
            self.assertIn("to", setting)
            self.assertIn("cc", setting)
            self.assertIn("notes", setting)

    def test_configure_grid(self):
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        view = self.make_view()

        # nb. avoid issues with arbitrary installed packages
        email_modules = [wuttaweb_emails]
        with patch.object(
            EmailHandler, "get_email_modules", return_value=email_modules
        ):
            grid = view.make_model_grid()
            self.assertIn("key", grid.searchable_columns)
            self.assertIn("subject", grid.searchable_columns)

    def test_render_to_short(self):
        view = self.make_view()
        setting = EmailSetting(self.config)

        # more than 2 recips
        result = view.render_to_short(
            setting,
            "to",
            [
                "alice@example.com",
                "bob@example.com",
                "charlie@example.com",
                "diana@example.com",
            ],
        )
        self.assertEqual(result, "alice@example.com, bob@example.com, ...")

        # just 2 recips
        result = view.render_to_short(
            setting,
            "to",
            [
                "alice@example.com",
                "bob@example.com",
            ],
        )
        self.assertEqual(result, "alice@example.com, bob@example.com")

        # just 1 recip
        result = view.render_to_short(setting, "to", ["alice@example.com"])
        self.assertEqual(result, "alice@example.com")

        # no recips
        result = view.render_to_short(setting, "to", [])
        self.assertIsNone(result)

    def test_get_instance(self):
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        view = self.make_view()

        # normal
        with patch.object(self.request, "matchdict", new={"key": "feedback"}):
            setting = view.get_instance()
            self.assertIsInstance(setting, dict)
            self.assertIn("key", setting)
            self.assertIn("sender", setting)
            self.assertIn("subject", setting)
            self.assertIn("to", setting)
            self.assertIn("cc", setting)
            self.assertIn("notes", setting)
            self.assertIn("enabled", setting)

        # not found
        with patch.object(
            self.request, "matchdict", new={"key": "this-should_notEXIST"}
        ):
            self.assertRaises(HTTPNotFound, view.get_instance)

    def test_get_instance_title(self):
        view = self.make_view()
        result = view.get_instance_title({"subject": "whatever"})
        self.assertEqual(result, "whatever")

    def test_configure_form(self):
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        view = self.make_view()

        with patch.object(self.request, "matchdict", new={"key": "feedback"}):
            setting = view.get_instance()
            form = view.make_model_form(setting)
            self.assertIn("description", form.readonly_fields)
            self.assertFalse(form.required_fields["replyto"])

    def test_persist(self):
        model = self.app.model
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        view = self.make_view()

        # start w/ no settings in db
        self.assertEqual(self.session.query(model.Setting).count(), 0)

        # "edit" settings for feedback email
        with patch.object(self.request, "matchdict", new={"key": "feedback"}):
            setting = view.get_instance()
            setting["subject"] = "Testing Feedback"
            setting["sender"] = "feedback@example.com"
            setting["replyto"] = "feedback4@example.com"
            setting["to"] = "feedback@example.com"
            setting["cc"] = "feedback2@example.com"
            setting["bcc"] = "feedback3@example.com"
            setting["notes"] = "did this work?"
            setting["enabled"] = True

            # persist email settings
            with patch.object(view, "Session", return_value=self.session):
                view.persist(setting)
                self.session.commit()

        # check settings in db
        self.assertEqual(self.session.query(model.Setting).count(), 8)
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.subject"),
            "Testing Feedback",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.sender"),
            "feedback@example.com",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.replyto"),
            "feedback4@example.com",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.to"),
            "feedback@example.com",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.cc"),
            "feedback2@example.com",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.bcc"),
            "feedback3@example.com",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.notes"),
            "did this work?",
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.enabled"), "true"
        )

        # "edit" settings for feedback email
        with patch.object(self.request, "matchdict", new={"key": "feedback"}):
            setting = view.get_instance()
            setting["subject"] = None
            setting["sender"] = None
            setting["replyto"] = None
            setting["to"] = None
            setting["cc"] = None
            setting["bcc"] = None
            setting["notes"] = None
            setting["enabled"] = False

            # persist email settings
            with patch.object(view, "Session", return_value=self.session):
                view.persist(setting)
                self.session.commit()

        # check settings in db
        self.assertEqual(self.session.query(model.Setting).count(), 1)
        self.assertIsNone(
            self.app.get_setting(self.session, "wutta.email.feedback.subject")
        )
        self.assertIsNone(
            self.app.get_setting(self.session, "wutta.email.feedback.sender")
        )
        self.assertIsNone(
            self.app.get_setting(self.session, "wutta.email.feedback.replyto")
        )
        self.assertIsNone(self.app.get_setting(self.session, "wutta.email.feedback.to"))
        self.assertIsNone(self.app.get_setting(self.session, "wutta.email.feedback.cc"))
        self.assertIsNone(
            self.app.get_setting(self.session, "wutta.email.feedback.bcc")
        )
        self.assertIsNone(
            self.app.get_setting(self.session, "wutta.email.feedback.notes")
        )
        self.assertEqual(
            self.app.get_setting(self.session, "wutta.email.feedback.enabled"), "false"
        )

    def test_render_to_response(self):
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        self.pyramid_config.add_route("home", "/")
        self.pyramid_config.add_route("login", "/auth/login")
        self.pyramid_config.add_route("email_settings", "/email/settings")
        self.pyramid_config.add_route(
            "email_settings.preview", "/email/settings/{key}/preview"
        )
        view = self.make_view()

        # nb. this gives coverage, but tests nothing..
        with patch.object(self.request, "matchdict", new={"key": "feedback"}):
            setting = view.get_instance()
            with patch.object(view, "viewing", new=True):
                context = {"instance": setting}
                response = view.render_to_response("view", context)
                self.assertIsInstance(response, Response)

    def test_preview(self):
        self.config.setdefault("wutta.email.default.sender", "test@example.com")
        view = self.make_view()

        # nb. this gives coverage, but tests nothing..
        with patch.object(self.request, "matchdict", new={"key": "feedback"}):

            # html
            with patch.object(self.request, "params", new={"mode": "html"}):
                response = view.preview()
                self.assertEqual(response.content_type, "text/html")

            # txt
            with patch.object(self.request, "params", new={"mode": "txt"}):
                response = view.preview()
                self.assertEqual(response.content_type, "text/plain")
